"""
A parent class that defines a common interface for all filter implementations.

Authors:
    | Tim J. Kuhlbusch <tim[dot]kuhlbusch[at]rwth-aachen.de>
    | Artem Basalaev <artem[dot]basalaev[at]pm.me>
"""
from typing import Type, TypeVar, Iterable, Optional
from warnings import warn
from copy import deepcopy
from pathlib import Path
import json
from dataclasses import dataclass, asdict

import numpy as np
from gwpy.timeseries import TimeSeriesDict

from spicypy.signal.time_series import TimeSeries

# create a type variable that can be any instance of a Filter subclass
FilterType = TypeVar("FilterType", bound="Filter")


@dataclass
class Filter:  # pylint: disable=too-many-instance-attributes
    """Parent class for filter implementations.
    Using this ensures a consistent interface for all filters and minimizes duplicate code.

    Function overloading by children must implement all non-optional parameters in the same position and order to guarantee compatibility.
    This is also enforced through a common test suite.

        Parameters
        ----------
        test:
            The signal that the filter will try to approximate.
        reference:
            Witness signal(s) to which the filter is applied.
        n_taps:
            Number of taps of the filter. This is equivalent to the number of time steps of the reference signal processed to get one prediction sample.
        n_proc: optional
            Number of parallel processes to use. Defaults to number of CPUs.
        subtract_mean: optional
            Subtract mean from data for internal processing, to ensure that floating point numbers are not too large. For the output time series operation is inverted, therefore besides improving numerical stability it is not affecting the results. Defaults to true.
        normalize_data: optional
            Divide data by its standard deviation for internal processing, to ensure that floating point numbers are not too large. For the output time series operation is inverted, therefore besides improving numerical stability it is not affecting the results. Defaults to true.
    """

    # The values defined here will be exported in the JSON dump
    # through the asdict() method and this being a dataclass.
    # Values in the parent class may not have default values, due to the way dataclasses work.
    test_mean: float
    test_std: float
    refs_mean: Iterable[float]
    refs_std: Iterable[float]
    reference_format: Optional[list]

    sample_rate: float
    n_references: int
    unit: str

    n_taps: int
    subtract_mean: bool
    normalize_data: bool

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        test: TimeSeries,
        reference: TimeSeriesDict | list,
        n_taps: int | None = None,
        subtract_mean: bool = True,
        normalize_data: bool = True,
        _from_dict: dict | None = None,
        **_kwargs,  # these will be ignored, except when
    ) -> None:
        # If _from_dict is passed, all other values will be ignored and the class will be initialized based on the dict content.

        self.subtract_mean = True
        self.normalize_data = True

        self.json_keys = [
            "sample_rate",
            "test_std",
            "test_mean",
            "refs_std",
            "refs_mean",
            "n_taps",
            "subtract_mean",
            "normalize_data",
            "n_references",
            "unit",
        ]
        self.optional_json_keys = [
            "reference_format",
        ]
        if hasattr(self, "additional_json_keys"):
            self.json_keys += list(self.additional_json_keys)

        if _from_dict is not None:
            self.loaded_from_file = True
            for key in self.json_keys:
                setattr(self, key, _from_dict.get(key))
            for key in self.optional_json_keys:
                if key in _from_dict:
                    setattr(self, key, _from_dict.get(key))
                else:
                    setattr(self, key, None)
            return
        else:
            self.loaded_from_file = False

        # The target time series.
        self.test = test
        # The witness signals.
        if isinstance(reference, TimeSeriesDict):
            self.reference_format = list(reference)
            self.reference = list(reference.values())
        elif isinstance(reference, list):
            self.reference_format = None
            self.reference = reference
        else:
            raise TypeError(
                "only TimeSeriesDict or list of TimeSeries are supported types for reference time series"
            )

        self.n_taps = n_taps or (len(self.test) - 1)
        self.subtract_mean = subtract_mean
        self.normalize_data = normalize_data

        # make a list of all inputs and compare values to test consistency
        inputs_list = self.reference + [self.test]
        self._check_timeseries_consistency(inputs_list)

        # copy time series parameters to filter properties
        self.sample_rate = self.test.sample_rate.value
        self.n_references = len(self.reference)
        self.unit = self.test.unit.to_string()

        self._check_mandatory_inputs()

    def _check_mandatory_inputs(self):
        """Check the input parameters for consistency."""
        if self.n_taps <= 0:
            raise ValueError(
                "You must specify a non-zero number of taps for the filter."
            )
        if self.n_taps >= 15000:
            warn(
                "Creating a filter with 15000 or more taps requested. This is A LOT. Depending on number of "
                "references and available resources, calculation may fail!"
            )
        if self.n_taps + 1 > len(self.test):
            warn(
                "Cannot set more taps than input time series length - 1. Setting n_taps to time series length - 1. "
            )
            self.n_taps = len(self.test) - 1

    def _check_timeseries_consistency(self, inputs_list):
        """Check input time series consistency (sampling rates, start time, length, etc.).

        Parameters
        ----------
        inputs_list: `list` of `TimeSeries`
            input time series
        """

        sample_rate = inputs_list[0].sample_rate.value
        epoch = inputs_list[0].epoch
        n_samples = len(inputs_list[0])

        # test time series for consistency
        for input_time_series in inputs_list:
            if len(input_time_series) != n_samples:
                raise ValueError(
                    "All time series must have the same length (number of samples)!"
                )
            if input_time_series.sample_rate.value != sample_rate:
                raise ValueError("All time series must have the same sampling rate!")
            if (
                input_time_series.epoch is not None
                and epoch is not None
                and input_time_series.epoch != epoch
            ):
                raise ValueError(
                    "All time series must be aligned in time (same time_series.epoch)!"
                )
            if np.iscomplexobj(input_time_series.value):
                raise NotImplementedError(
                    "Filtering for complex time series is not implemented."
                )

    def _prepare_data(self):
        """
        Normalize input time series to values close to 1 if requested, for better numerical stability.
        Subtract mean if requested.
        """

        self.test_array = deepcopy(self.test.value)  # need to copy time series
        # or else normalization will affect original values

        self.test_mean = np.mean(self.test_array)
        self.test_std = np.std(self.test_array)

        self.reference_array = np.array(self.reference, copy=True).T
        self.refs_mean = np.mean(self.reference_array, axis=0)
        self.refs_std = np.std(self.reference_array, axis=0)

        if self.subtract_mean:
            self.test_array -= self.test_mean
            self.reference_array -= self.refs_mean
        if self.normalize_data:
            self.test_array /= self.test_std
            self.reference_array /= self.refs_std

    @staticmethod
    def _make_json_filname(filename: str) -> Path:
        """Append the `.json` suffix to the given filename if it does not already have it."""
        # convert to Path object and append file ending if necessary
        path = Path(filename)
        if path.suffix != ".json":
            path = Path(filename + ".json")
        return path

    def save(self, filename: str):
        """Save the filter as a JSON text file.

        Parameters
        ----------
        filename:
            Target filename. If file name does not end in `.json`, it will be appended.
        """
        with open(self._make_json_filname(filename), "w", encoding="utf-8") as out_file:
            filter_dict = self.as_dict()
            filter_dict = {
                k: (v.tolist() if isinstance(v, np.ndarray) else v)
                for k, v in filter_dict.items()
            }
            json.dump(filter_dict, out_file, default=str)

    @classmethod
    def load(cls: Type[FilterType], filename: str) -> FilterType:
        """Load the filter from a JSON text file.

        Parameters
        ----------
        filename: `str`
            Target filename. If file name does not end on `.json`, it will be appended.

        Returns
        -------
        New filter object.
        """
        with open(
            Filter._make_json_filname(filename), "r", encoding="utf-8"
        ) as in_file:
            data_dict = json.load(in_file)

        # All other arguments are ignored if _from_dict is supplied.
        # The values here are just placeholders to match the signature of the user facing inializer.
        return cls(TimeSeries([]), [], n_taps=1, _from_dict=data_dict)

    def as_dict(self) -> dict:
        """Return a JSON compatible dict representation of the filter."""
        return asdict(self)

    def create_filters(self) -> None:  # TODO: this does not work as a generic name!
        """A function that must be called by the user after instantiating the class. This function conditions the filter based on the input provided during instantiation."""
        raise NotImplementedError(
            "create_filters() must be implemented by child class."
        )

    def _preprocess_data(self, to_training_data, inputs_list):
        """Preprocess data (check, normalize, subtract mean, etc) before applying the filter.

        Parameters
        ----------
        to_training_data: `bool`
            Whether the filter is applied to training data.
        inputs_list: `np.array` of `float`
            Array of inputs (references) to which this filter is applied.

        Returns
        -------
        refs: `Iterable[np.array]`
            The prepared reference signal.
        """

        if (inputs_list is None and not to_training_data) or (
            inputs_list is not None and to_training_data
        ):
            raise ValueError(
                "Either a reference time series must be specified or 'to_training_data' set to 'True', but not both."
            )

        if to_training_data:
            return self.reference_array

        if self.n_taps >= len(inputs_list[0]):
            raise ValueError(
                "Input time series has less time steps than number of filter taps! Cannot apply filter."
            )

        # check consistency
        self._check_timeseries_consistency(inputs_list)
        if inputs_list[0].sample_rate.value != self.sample_rate:
            raise ValueError(
                f"Reference sample rate is not the same as sample rate used to generate this filter (expected {self.sample_rate}, got {inputs_list[0].sample_rate.value})!"
            )

        if len(inputs_list) != self.n_references:
            raise ValueError("Wrong number of references for this Filter")

        refs = np.array(inputs_list, copy=True).T

        for i in range(refs.shape[1]):
            if self.normalize_data:
                refs[:, i] /= self.refs_std[i]
            if self.subtract_mean:
                refs[:, i] -= self.refs_mean[i]
        return refs

    @staticmethod
    def _check_apply_parameters(func):
        """Decorator for create_filters() impelementations that checks the mandatory parameters.

        Converts inputs_list from TimeSeriesDict to list.
        """

        def wrapper(self, inputs_list=None, *args, **kwargs):
            to_training_data = kwargs.get("to_training_data") or (
                args[2] if len(args) > 2 else False
            )

            if to_training_data and self.loaded_from_file:
                raise NotImplementedError(
                    "Cannot apply filter to training data because this filter "
                    "was loaded from file; filters are already applied to training data earlier. "
                    "Try calling `.apply(..)` without `to_training_data` argument."
                )

            # if the reference values given at apply were a TimeSeriesDict
            if not to_training_data:
                if self.reference_format is None:
                    if isinstance(inputs_list, TimeSeriesDict):
                        raise ValueError(
                            "This filter was not initialized from a TimeSeriesDict and does not accept a TimeSeriesDict as inputs_list."
                        )
                else:
                    if not isinstance(inputs_list, TimeSeriesDict):
                        raise ValueError(
                            "This filter was initialized from a TimeSeriesDict and only accepts TimeSeriesDict for inputs_list."
                        )

                    inputs_list = [inputs_list[key] for key in self.reference_format]

            return func(self, inputs_list, *args, **kwargs)

        return wrapper

    @_check_apply_parameters
    def apply(
        self,
        inputs_list: Optional[Iterable[TimeSeries]] = None,
        zero_padding: bool = True,
        normalized_output: bool = False,
        to_training_data: bool = False,
        use_multiprocessing: bool = False,
    ):  # pylint: disable=too-many-positional-arguments
        """Apply the filter to given input data.
        inputs_list: `list` of `TimeSeries` or `TimeSeriesDict`, optional
            time series to apply filter to, not needed if applying to training data
        zero_padding: `bool`, optional
            if true, return time series of the same length as inputs,
            with zeros in the beginning (filter needs to "accumulate" data equal to its n_taps first);
            if false, returned time series will be shorter and with shifted t0 by n_taps/sample_rate
        normalized_output: `bool`, optional
            if true, do not add mean and multiply by standard deviation, as if the data were already normalized
        to_training_data: `bool`, optional
            apply filters to training data - runs after filter creation to de3termine best filter,
            but can also be called explicitly
        use_multiprocessing: `bool`, optional
            Whether to use multiprocessing (with number of parallel processes set with `self.n_proc`)

        Returns
        -------
        prediction :  `TimeSeries`
            filter output
        """
        raise NotImplementedError("apply() must be implemented by child class.")



"""
Class calculating transfer function between two time series (result is FrequencySeries).

Authors:
    | Artem Basalaev <artem[dot]basalaev[at]pm.me>
    | Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
    | Abhinav Patra <patraa1[at]cardiff.ac.uk>
    | Octavio Vega <vega00087[at]gmail[dot]com>
"""
from typing import Optional

import gwpy
import numpy as np
import matplotlib

from spicypy.signal.frequency_series import FrequencySeries
from spicypy.signal.time_series import TimeSeries


class TransferFunction:  # pylint: disable=too-many-instance-attributes
    """
    Class calculating transfer function between two time series (result is FrequencySeries)

    """

    def __init__(
        self,
        ts1: Optional[TimeSeries] = None,
        ts2: Optional[TimeSeries] = None,
        tf: Optional[FrequencySeries] = None,
        name="Generic transfer function",
        **kwargs,
    ):
        """Constructor takes either two time series _or_ another TransferFunction object

        Parameters
        ----------
        ts1 : `TimeSeries`
            first `TimeSeries`
        ts2 : `TimeSeries`
            second `TimeSeries`
        tf : `TransferFunction`
            alternative constructor with another TransferFunction object
        name : `str`, optional
            name, if not provided constructed from input time series names
        kwargs : `dict`, optional
            any additional arguments that `TimeSeries.psd()` or `TimeSeries.csd()` can take, e.g. averaging method
        """
        self._ts1 = ts1
        self._ts2 = ts2
        self._kwargs = kwargs
        self._coherence = None
        self._psd1 = None
        self._psd2 = None
        self._csd = None
        self._tf = tf
        self._name = name

        # if we specify the transfer function directly (nothing to calculate)
        if self._tf is not None:
            # check that other args are not specified
            if ts1 is not None or ts2 is not None:
                raise ValueError(
                    "Transfer function was specified on input together with time series"
                )
            self._tf.name = name
            return

        if ts1 is None or ts2 is None:
            raise ValueError("Two time series should be specified on input (got <2)")

    @property
    def ts1(self):
        """Property holding first time series

        Returns
        -------
        psd1 :  TimeSeries
            first time series
        """
        return self._ts1

    @property
    def ts2(self):
        """Property holding second time series

        Returns
        -------
        psd1 :  TimeSeries
            second time series
        """
        return self._ts2

    @property
    def psd1(self):
        """Calculate the PSD of the first time series

        Returns
        -------
        psd1 :  FrequencySeries
            a data series containing the PSD.
        """
        if self._psd1 is None:
            self._psd1 = self._ts1.psd(**self._kwargs)
        return self._psd1

    @property
    def psd2(self):
        """Calculate the PSD of the second time series

        Returns
        -------
        psd2 :  FrequencySeries
            a data series containing the PSD.
        """
        if self._psd2 is None:
            self._psd2 = self._ts2.psd(**self._kwargs)
        return self._psd2

    @property
    def csd(self):
        """Calculate the CSD of the first time series to the second

        Returns
        -------
        psd1 :  FrequencySeries
            a data series containing the CDD.
        """
        if self._csd is None:
            self._csd = self._ts1.csd(self._ts2, **self._kwargs)
        return self._csd

    @property
    def tf(self):
        """Calculate the transfer function

        Returns
        -------
        tf :  FrequencySeries
            a data series containing the transfer function.
        """
        if self._tf is None:
            self._tf = self._ts1.transfer_function(self._ts2, **self._kwargs)
            # if we didn't already specify a nice name for it...
            if self._name == "Generic transfer function":
                self.tf.name = (
                    f"Transfer Function between {self._ts1.name} and {self._ts2.name}"
                )
        return self._tf

    @property
    def coherence(self):
        """Calculate coherence between input time series

        Returns
        -------
        coherence :  FrequencySeries
            a data series containing normalized coherence in frequency domain
        """
        if self._coherence is None:
            self._coherence = self._ts1.coherence(self._ts2, **self._kwargs)
        return self._coherence

    def plot(self, coherence=False, **kwargs) -> matplotlib.figure.Figure:
        """Plot transfer function (magnitude and phase) and optionally coherence

        Returns
        -------
        plot :  :class:`matplotlib.figure.Figure`
            a matplotlib plot that can be further modified by the user
        """
        if self.tf is None:
            raise AttributeError(
                "Somehow missing transfer function, this should never happen"
            )
        if self._ts1 is None or self._ts2 is None:
            return self.tf.plot(**kwargs)

        traces = [np.abs(self.tf), np.angle(self.tf, deg=True)]
        if coherence:
            traces.append(self.coherence)

        # log axis by default
        xscale = kwargs.pop("xscale", None)
        if xscale is None:
            xscale = "log"
        yscale = kwargs.pop("yscale", None)
        if yscale is None:
            yscale = "log"

        plot = gwpy.plot.Plot(
            *traces,
            separate=True,
            figsize=(10, 10),
            xscale=xscale,
            yscale=yscale,
            **kwargs,
        )
        plot.align_labels()

        # Magnitude of transfer function
        plot.axes[0].set_title(self.tf.name)
        if self._ts1.unit == self._ts2.unit:
            plot.axes[0].set_ylabel("Magnitude")
        else:
            plot.axes[0].set_ylabel(self._ts2.unit / self._ts1.unit)
        # Phase of transfer function
        plot.axes[1].set_ylabel("Phase [deg]")
        plot.axes[1].set_yscale("linear")

        if coherence:
            # Coherence between input time series
            plot.axes[2].set_title(traces[2].name)
            plot.axes[2].set_ylabel("Coherence")
            plot.axes[2].set_yscale("linear")

            if np.all(np.isclose(traces[2].value, 1.0, rtol=0.99)):
                plot.axes[2].set_ylim(0.7, 1.3)

        plot.tight_layout()

        return plot



"""
===============
Coherent subtraction (Huddle test in presence of signals)
===============

| Modified for Spicypy by Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Reviewed by Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
| Contributors:
| Shreevathsa Chalathadka Subrahmanya <schalath[at]physnet.uni-hamburg.de>

Original description:

An attempt to make mccs in Python. We generalise this to support cross spectral
densities, something I assert is useful.

Although I only, properly, understand an alternative method for doing this
procedure that suffers from poor behaviour with non-linearly independent noise
sources. Alternatively I know that mccs works, and the community is happy with
its performance - so I will copy its functionality.

| Authors: Nathan A. Holland, Pooya Saffarieh, Brian Lantz, Conor Mow-Lowry
| Contact: nholland@nikhef.nl
| Date: 2022-05-20

Improvements to make:
    - Error checking needs to be improved
    - Support for more than 1 dimension in <x> and <y> data, and more than 2 dimensions for <ref> data - this was too hard for me to devise on the first programming of this function
    - Adjusted scaling factor for first bin (DC bin)
    - Better scaling factor for more than 1 ref. channel
    - Add warning for rank issues in lstsq
    - Return warning when input data is too small for a given average length

Major Difference from mccs2 in matlab:
    - We don't put zero in the first bin (DC)
    - For even number of data points here it returns n/2 + 1 rather than n/2

Pending important tests:
    - Non-hermitian signals.
    - Other window function rather than Hann and Rectangular.

"""

import numpy as np
from scipy.linalg import lstsq
from scipy.fft import fft, rfft, fftfreq, rfftfreq

from gwpy.timeseries import TimeSeriesDict

from spicypy.signal.spectral import (
    AveragingParameters,
    daniell_rearrange_fft,
    daniell_average,
)


def _coherent_subtract_psd(bundle):
    """Calculate a single coherently subtracted PSD for a spectrogram.

    Parameters
    ----------
    bundle : `tuple`
        test: `TimeSeries`
            input time series
        ref: list or `TimeSeriesDict`
            reference time series
        kwargs : `dict`
            additional arguments for Daniell method averaging

    Returns
    -------
    PSD : FrequencySeries
        residual spectrum PSD
    """
    test, ref, kwargs = bundle
    psd_ = coherent_subtraction(test, ref, **kwargs)
    return psd_


def coherent_subtraction_spectrogram(test, reference, stride, **kwargs):
    """Coherent subtraction spectrogram of reference time series from test time series (typical use case: Huddle test)

    Parameters
    ----------
    test: `TimeSeries`
        input time series
    reference: list or  `TimeSeriesDict`
        reference time series
    stride: `float`
        stride for spectrogram, in seconds
    kwargs : `dict`
        additional arguments for Daniell method averaging

    Returns
    -------
    Spectrogram : Spectrogram
        residual spectrogram
    """
    #  enable multiprocessing for spectrogram
    from gwpy.utils import mp
    from .spectrogram import Spectrogram

    # process input references
    reference_list = reference
    if isinstance(reference, TimeSeriesDict):
        reference_list = []
        for channel in reference.keys():
            reference_list.append(reference[channel])
    elif not isinstance(reference, list):
        raise TypeError(
            "only TimeSeriesDict or list of TimeSeries are supported types for reference time series"
        )

    # chunk time series
    test_slices = test.chunk(stride)
    sliced_refs = []  # each element here is single reference, all chunks
    for ref in reference_list:
        sliced_refs.append(ref.chunk(stride))

    n_chunks = len(sliced_refs[0])
    n_refs = len(reference_list)
    # swap the lists and form inputs for each PSD calculation
    inputs = [
        (test_slices[i], [sliced_refs[j][i] for j in range(n_refs)], kwargs)
        for i in range(n_chunks)
    ]

    # calculate spectrogram
    psds = mp.multiprocess_with_queues(
        1, _coherent_subtract_psd, inputs
    )  # nproc=1 hardcoded, but so is in GWpy
    epoch = test.t0.value
    return Spectrogram.from_spectra(*psds, epoch=epoch, dt=stride)


def coherent_subtraction(*args, **kwargs):
    """Coherent subtraction of reference time series from test time series (typical use case: Huddle test)

    Parameters
    ----------
    args : `tuple`
        timeseries: `TimeSeries`
            input time series (if only one specified, calculates PSD)
        other: `TimeSeries`, optional
            second input time series
        reference: list or  `TimeSeriesDict`
            reference time series
    kwargs : `dict`
        additional arguments for Daniell method averaging

    Returns
    -------
    CSD : FrequencySeries
        residual spectrum (either CSD for two test inputs, of PSD for one)
    """

    try:
        test1, reference = args
        test2 = test1
        is_csd = False
    except ValueError:
        test1, test2, reference = args
        is_csd = True

    reference_list = reference
    if isinstance(reference, TimeSeriesDict):
        reference_list = []
        for channel in reference.keys():
            reference_list.append(reference[channel])
    elif not isinstance(reference, list):
        raise TypeError(
            "only TimeSeriesDict or list of TimeSeries are supported types for reference time series"
        )

    cs = CoherentSubtraction(test1, test2, reference_list, is_csd, **kwargs)
    # process input arguments
    cs.check_inputs()
    # calculate FFTs for all inputs
    cs.prepare_ffts()
    # perform coherent subtraction
    cs.subtract()
    # average frequency bins, construct and normalize CSD, return
    return cs.average_CSD()


class CoherentSubtraction:  # pylint: disable=too-many-instance-attributes
    """
    Class performing coherent subtraction in frequency domain
    """

    def __init__(self, test1, test2, reference, is_csd, **kwargs):
        """Init method for coherent subtraction

        Parameters
        ----------
        test1: `TimeSeries`
            input time series (if only one specified, calculates PSD)
        test2: `TimeSeries`, optional
            second input time series
        reference: list or  `TimeSeriesDict`
            reference time series
        is_csd: bool
            a flag indicating whether subtraction is from aCSD or a PSD
        kwargs : `dict`
            additional arguments for Daniell method averaging
        """
        self.test1 = test1
        self.test2 = test2
        self.reference = reference
        self.is_csd = is_csd
        self.kwargs = kwargs

        # and set some defaults
        self.is_onesided = True
        self.detrend = "constant"
        self.ap = None
        self.test1_fft = None
        self.test2_fft = None
        self.reference_fft = None
        self.frq = None
        self.CSD = None
        self.bin_widths = None

    def check_inputs(self):
        """Check input arguments for coherent subtraction. Daniell-averaging-method-specific arguments are checked by `AveragingParameters` class."""
        # make a list of all other inputs and compare values to test1
        inputs_list = self.reference + [self.test1, self.test2]
        sample_rate = self.test1.sample_rate.value
        epoch = self.test1.epoch
        is_onesided = True

        for input_time_series in inputs_list:
            if len(input_time_series) != len(self.test1):
                raise ValueError(
                    "All time series must have the same length (number of samples)!"
                )
            if input_time_series.sample_rate.value != sample_rate:
                raise ValueError("All time series must have the same sampling rate!")
            if (
                input_time_series.epoch is not None
                and epoch is not None
                and input_time_series.epoch != epoch
            ):
                raise ValueError(
                    "All time series must be aligned in time (same time_series.epoch)!"
                )
            if np.iscomplexobj(input_time_series.value):
                is_onesided = (
                    False  # if any inputs are complex, will have two-sided fft
                )

        self.detrend = self.kwargs.pop("detrend", "constant")
        if self.detrend != "constant" and self.detrend != "linear":
            raise NotImplementedError(
                "Can only apply 'constant' and 'linear' detrend to time series"
            )

        # get parameters for Daniell method
        self.ap = AveragingParameters(sample_rate, len(self.test1), **self.kwargs)
        if len(self.reference) >= self.ap.number_averages // 2 + 1:
            raise ValueError(
                "For coherent subtraction to work, number of references must be <= number_averages // 2 + 1."
                "Try to increase `number_averages` value"
            )
        self.ap.is_onesided = is_onesided

    def prepare_ffts(self):
        """Detrend time series and calculate FFTs. Then rearrange ffts for Daniell averaging methods using `daniell_rearrange_fft` function."""
        # one-sided or two-sided fft
        if self.ap.is_onesided:
            fft_func = rfft
            fftfreq_func = rfftfreq
        else:
            fft_func = fft
            fftfreq_func = fftfreq

        # detrend time series, apply window, and perform fft (full length of time series is used, no segmenting)
        test1_fft = fft_func(
            self.test1.detrend(self.detrend).value * self.ap.win, self.ap.nfft
        )
        test2_fft = fft_func(
            self.test2.detrend(self.detrend).value * self.ap.win, self.ap.nfft
        )
        reference_fft = []
        for ref in self.reference:
            reference_fft.append(
                fft_func(ref.detrend(self.detrend).value * self.ap.win, self.ap.nfft)
            )

        # calculate corresponding frequencies
        frq = fftfreq_func(self.ap.nfft, 1.0 / self.ap.sample_rate)

        # now rearrange ffts according to Daniell method
        self.frq, self.test1_fft, self.ap = daniell_rearrange_fft(
            frq, test1_fft, self.ap
        )
        _, self.test2_fft, _ = daniell_rearrange_fft(frq, test2_fft, self.ap)
        self.reference_fft = []
        for ref_fft in reference_fft:
            _, fft_rearranged, _ = daniell_rearrange_fft(frq, ref_fft, self.ap)
            self.reference_fft.append(fft_rearranged)

    def subtract(self):
        """Perform coherent subtraction in frequency domain. Main method of the algorithm"""
        # perform coherent subtraction
        n_samples = len(self.test1_fft)
        number_reference_channels = len(self.reference_fft)
        self.CSD = np.full(n_samples, np.nan, dtype=complex)
        edges, self.bin_widths = self.ap.window_edges(self.CSD)
        left_edge = 0
        # do least square fit in each bin
        for right_edge in edges:
            test1_bin = self.test1_fft[left_edge:right_edge]
            test2_bin = self.test2_fft[left_edge:right_edge]
            ref_bin = np.zeros(
                [len(test1_bin), number_reference_channels], dtype=complex
            )

            # temporarily store references in another array to rearrange (see below)
            ref_tmp_bin = []
            for ref_fft in self.reference_fft:
                ref_tmp_bin.append(ref_fft[left_edge:right_edge])
            # Swap element indices for matrix operations on the next step to work
            # turns [[ref1_f1, ref1_f2, ref1_f3], [ref2_f1, ref2_f2, ref2_f3]]
            # to [[ref1_f1, ref2_f1] [ref1_f2, ref2_f2], [ref1_f3, ref2_f3]]
            for t in range(len(test1_bin)):
                for k in range(number_reference_channels):
                    ref_bin[t][k] = ref_tmp_bin[k][t]

            # turns [test1_f1, test1_f2, test1_f3]
            # to [[test1_f1], [test1_f2], [test1_f3]]
            test1_bin = test1_bin.reshape(len(test1_bin), -1)
            test2_bin = test2_bin.reshape(len(test2_bin), -1)

            # Get best fit coefficients, test_fit ,from least-square fit of references,
            # ref_bin, to test, test_bin. (And same for another test signal _Y)
            test1_fit, _, _, _ = lstsq(ref_bin, test1_bin)
            test2_fit, _, _, _ = lstsq(ref_bin, test2_bin)

            # add some warning for rank being less than the channels.
            # lstsq returns array size zero when channels are linear combination of each other.
            # so we calculate residuals manually here

            # calculating residuals
            test1_res = test1_bin - np.matmul(ref_bin, test1_fit)
            test2_res = test2_bin - np.matmul(ref_bin, test2_fit)
            # multiplying residuals = forming the CSD
            # flattening because array is still technically 2D, with dimensions 1 x number_averages
            self.CSD[left_edge:right_edge] = np.ndarray.flatten(
                test1_res * test2_res.conj()
            )
            if self.ap.is_onesided:
                # in one-sided case, scale factor 2x should be applied to properly normalize CSD
                self.CSD[left_edge:right_edge] *= 2.0
            left_edge = right_edge

    def average_CSD(self):
        """Construct CSD and average using Daniell averaging method (`daniell_average`)"""
        # do correct scaling for density or spectrum
        if self.ap.scaling == "density":
            self.CSD /= self.ap.sample_rate * (self.ap.win * self.ap.win).sum()
        else:  # can only be "spectrum" - this argument already sanitized by Daniell method
            self.CSD /= self.ap.win.sum() ** 2
        # and override scale argument for Daniell method to prevent double scaling
        self.ap.scale = 1.0
        # also override bias_correction (different because we have multiple reference channels)
        self.ap.set_bias_correction(
            self.ap.bias_correction(self.CSD)
            - len(self.reference_fft) / np.array(self.bin_widths)
        )

        # remove imaginary part if it's zero
        if not np.iscomplex(self.CSD).any():
            self.CSD = np.real(self.CSD)

        # propagate parameters from time series
        timeseries = self.test1
        if self.is_csd:
            if self.test1.name is not None and self.test2.name is not None:
                timeseries.name = (
                    "Coherent subtraction from CSD("
                    + self.test1.name
                    + ","
                    + self.test2.name
                    + ")"
                )
            else:
                timeseries.name = "Coherent subtraction from CSD"
        else:
            if self.test1.name is not None:
                timeseries.name = (
                    "Coherent subtraction from PSD(" + self.test1.name + ")"
                )
            else:
                timeseries.name = "Coherent subtraction from PSD"
        # and average
        return daniell_average(self.frq, self.CSD, timeseries, self.ap)



"""
Class extending functionality of :obj:`gwpy.timeseries.timeseries.TimeSeries` from GWpy.

Authors:
    | Artem Basalaev <artem[dot]basalaev[at]pm.me>
    | Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
    | Abhinav Patra <patraa1[at]cardiff.ac.uk>
    | Octavio Vega <vega00087[at]gmail[dot]com>
    | Jonathan Perry <j.w.perry[at]vuDOTnl>
"""
from functools import wraps
import copy
from warnings import warn
import numpy as np
import gwpy.timeseries
import gwpy.frequencyseries
import gwpy.spectrogram
from gwpy.signal import spectral

from spicypy.signal.spectral import daniell, lpsd
from spicypy.signal.coherent_subtraction import (
    coherent_subtraction,
    coherent_subtraction_spectrogram,
)

spectral.register_method(daniell)
spectral.register_method(lpsd)


def check_not_implemented_method(func):
    """Decorate a method to check if user tries to do LPSD or Daniell averaging method
    where we haven't implemented it yet.
    """

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        """Wrap method to check if user tries to do LPSD or Daniell averaging method
        where we haven't implemented it yet.
        """
        if "method" in kwargs and (
            kwargs["method"] == "daniell" or kwargs["method"] == "lpsd"
        ):
            raise NotImplementedError(
                "LPSD/Daniell averaging is not supported for this method"
            )
        return func(*args, **kwargs)

    return wrapped_func


def fix_window_argument(func):
    """Decorate a method to fix "window" argument for Spicypy"""

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        """Wrap method to fix "window" argument for Spicypy"""
        # work-around to propagate 'window' argument intact for custom averaging methods
        window = kwargs.get("window")
        if "method" in kwargs and (
            kwargs["method"] == "daniell" or kwargs["method"] == "lpsd"
        ):
            kwargs["window_"] = window
        elif window is None:
            # default to 'hann' for standard GWpy methods
            kwargs["window"] = "hann"

        return func(*args, **kwargs)

    return wrapped_func


def fix_return_object(func):
    """Decorate a method to always return Spicypy object if available"""

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        """Wrap method to always return Spicypy object if available"""
        res = func(*args, **kwargs)
        # convert to Spicypy object
        if isinstance(res, gwpy.frequencyseries.frequencyseries.FrequencySeries):
            from spicypy.signal import FrequencySeries

            return FrequencySeries.from_other(res)
        elif isinstance(res, gwpy.timeseries.timeseries.TimeSeries):
            return TimeSeries.from_other(res)
        elif isinstance(res, gwpy.spectrogram.Spectrogram):
            from spicypy.signal import Spectrogram

            return Spectrogram.from_other(res)
        return res

    return wrapped_func


def _psd(bundle):
    """Calculate a PSD for a spectrogram.

    Parameters
    ----------
    bundle : `tuple`
        ts: `TimeSeries`
            input time series
        kwargs : `dict`
            additional arguments for PSD

    Returns
    -------
    psd_ : FrequencySeries
        a PSD of these time series
    """
    ts, kwargs = bundle
    psd_ = ts.psd(**kwargs)
    return psd_


class TimeSeries(gwpy.timeseries.TimeSeries):
    """
    Class to model signals (time series)

    """

    @classmethod
    def from_other(cls, ts):
        """Create TimeSeries from another TimeSeries, in particular gwpy.TimeSeries

        Parameters
        ----------
        ts : TimeSeries, gwpy.timeseries.TimeSeries
            Other TimeSeries object
        Returns
        -------
        TimeSeries
            Copy Spicypy TimeSeries object
        """
        ts_spicypy = cls(ts.value)
        ts_spicypy.__dict__ = copy.deepcopy(ts.__dict__)
        return ts_spicypy

    @fix_window_argument
    @fix_return_object
    def asd(self, fftlength=None, overlap=None, window=None, **kwargs):
        """Calculate the ASD `FrequencySeries` of this `TimeSeries`

        Parameters
        ----------
        fftlength : `float`, optional
            number of seconds in single FFT. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, calculates single FFT covering full duration and then performs averaging in frequency domain.
            * For LPSD averaging method (`method='lpsd'`): user-specified value ignored, algorithm calculates optimal segment lengths.
            * For other averaging methods: defaults to a single FFT covering the full duration

        overlap : `float`, optional
            number of seconds of overlap between FFTs. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, no overlap possible because a single FFT is calculated.
            * For other averaging methods: defaults to the recommended overlap for the given window (if given), or 0

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. Behavior depends on averaging method:

            * For LPSD averaging method (`method='lpsd'`): only `str` type is allowed. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
            * For other averaging methods: see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method : `str`, optional
                FFT-averaging method (default: ``'median'``). The accepted ``method`` arguments are:

                - ``'bartlett'`` : a mean average of non-overlapping periodograms
                - ``'median'`` : a median average of overlapping periodograms
                - ``'welch'`` : a mean average of overlapping periodograms
                - ``'lpsd'`` :  average of overlapping periodograms binned logarithmically in frequency
                - ``'daniell'`` : calculates single fft for the whole time series and averages in frequency domain

            any other keyword arguments accepted by the respective averaging methods. See definitions of corresponding method (`method` keyword). If `method` is not specified, defaults to :class:`gwpy.signal.spectral.csd`

        Returns
        -------
        asd :  FrequencySeries
            a data series containing the ASD
        """
        return super().asd(fftlength, overlap, window, **kwargs)

    @fix_window_argument
    @fix_return_object
    def psd(self, fftlength=None, overlap=None, window=None, **kwargs):
        """Calculate the PSD `FrequencySeries` of this `TimeSeries`

        Parameters
        ----------
        fftlength : `float`, optional
            number of seconds in single FFT. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, calculates single FFT covering full duration and then performs averaging in frequency domain.
            * For LPSD averaging method (`method='lpsd'`): user-specified value ignored, algorithm calculates optimal segment lengths.
            * For other averaging methods: defaults to a single FFT covering the full duration

        overlap : `float`, optional
            number of seconds of overlap between FFTs. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, no overlap possible because a single FFT is calculated.
            * For other averaging methods: defaults to the recommended overlap for the given window (if given), or 0

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. Behavior depends on averaging method:

            * For LPSD averaging method (`method='lpsd'`): only `str` type is allowed. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
            * For other averaging methods: see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method : `str`, optional
                FFT-averaging method (default: ``'median'``). The accepted ``method`` arguments are:

                - ``'bartlett'`` : a mean average of non-overlapping periodograms
                - ``'median'`` : a median average of overlapping periodograms
                - ``'welch'`` : a mean average of overlapping periodograms
                - ``'lpsd'`` :  average of overlapping periodograms binned logarithmically in frequency
                - ``'daniell'`` : calculates single fft for the whole time series and averages in frequency domain

            any other keyword arguments accepted by the respective averaging methods.
            See definitions of corresponding method (`method` keyword). If `method` is not specified,
            defaults to gwpy.signal.spectral.csd

        Returns
        -------
        psd :  FrequencySeries
            a data series containing the PSD
        """
        return super().psd(fftlength, overlap, window, **kwargs)

    @fix_window_argument
    @fix_return_object
    def csd(self, other, fftlength=None, overlap=None, window=None, **kwargs):
        """Calculate the CSD `FrequencySeries` for two `TimeSeries`

        Parameters
        ----------
        other : `TimeSeries`
            the second `TimeSeries` in this CSD calculation

        fftlength : `float`, optional
            number of seconds in single FFT. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, calculates single FFT covering full duration and then performs averaging in frequency domain.
            * For LPSD averaging method (`method='lpsd'`): user-specified value ignored, algorithm calculates optimal segment lengths.
            * For other averaging methods: defaults to a single FFT covering the full duration

        overlap : `float`, optional
            number of seconds of overlap between FFTs. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, no overlap possible because a single FFT is calculated.
            * For other averaging methods: defaults to the recommended overlap for the given window (if given), or 0

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. Behavior depends on averaging method:

            * For LPSD averaging method (`method='lpsd'`): only `str` type is allowed. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
            * For other averaging methods: see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method: `str`, optional
                averaging method for coherence calculation (default: ``'median'``). See above for important difference in arguments. The accepted ``method`` arguments are:

                - ``'bartlett'`` : a mean average of non-overlapping periodograms
                - ``'median'`` : a median average of overlapping periodograms
                - ``'welch'`` : a mean average of overlapping periodograms
                - ``'lpsd'`` :  average of overlapping periodograms binned logarithmically in frequency
                - ``'daniell'`` : calculates single fft for the whole time series and averages in frequency domain

            any other keyword arguments accepted by the respective averaging methods.
            See definitions of corresponding method (`method` keyword).

        Returns
        -------
        csd :  FrequencySeries
            a data series containing the CSD.
        """

        method_func = spectral.csd
        method = kwargs.pop("method", None)
        if method == "daniell":
            method_func = daniell
        elif method == "lpsd":
            method_func = lpsd
        elif method is None:
            # using default GWpy method; in that case, default fftlength will may also be used
            # inform the user of dangers
            if fftlength is None:
                warn(
                    "No 'fftlength' specified, note that in this case single FFT covering whole time series is used"
                )
        else:
            raise NotImplementedError(
                "Only 'daniell' and 'lpsd' averaging methods are currently implemented in addition to default"
            )

        return spectral.psd(
            (self, other),
            method_func,
            fftlength=fftlength,
            overlap=overlap,
            window=window,
            **kwargs,
        )

    def coherent_subtract_spectrogram(self, reference, stride, window=None, **kwargs):
        """Calculate the residual spectrogram (using coherent subtraction) between this `TimeSeries` and references.

        Parameters
        ----------
        reference: list or  `TimeSeriesDict`
            reference time series

        stride: `float`
            stride for spectrogram, in seconds

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT.

            see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method: `str`, optional
                averaging method for coherence calculation. Currently only 'daniell' is supported

            any other keyword arguments accepted by 'daniell' averaging method

        Returns
        -------
        residual_spectrogram : Spectrogram
            residual spectrogram after frequency-domain subtraction of reference time series
        """

        method = kwargs.pop("method", "daniell")
        if method != "daniell":
            raise NotImplementedError(
                "Coherent subtraction is currently implemented only with Daniell averaging "
                "method."
            )
        if kwargs.get("overlap"):
            raise ValueError(
                "overlap argument is not supported by Daniell averaging method; "
                "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
            )
        if kwargs.get("fftlength"):
            raise ValueError(
                "fftlength argument is not supported by Daniell averaging method; "
                "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
            )
        kwargs["window_"] = window
        return coherent_subtraction_spectrogram(self, reference, stride, **kwargs)

    def coherent_subtract(self, reference, window=None, **kwargs):
        """Calculate the residual PSD (using coherent subtraction) between this `TimeSeries` and references.

        Parameters
        ----------
        reference: list or  `TimeSeriesDict`
            reference time series

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT.

            see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method: `str`, optional
                averaging method for coherence calculation. Currently only 'daniell' is supported

            any other keyword arguments accepted by 'daniell' averaging method

        Returns
        -------
        residual_psd : FrequencySeries
            residual PSD after frequency-domain subtraction of reference time series
        """

        # No overlap for Daniell method.
        if kwargs.get("overlap"):
            raise ValueError(
                "overlap argument is not supported by Daniell averaging method; "
                "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
            )
        if kwargs.get("fftlength"):
            raise ValueError(
                "fftlength argument is not supported by Daniell averaging method; "
                "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
            )
        kwargs["window_"] = window
        return coherent_subtraction(self, reference, **kwargs)

    @fix_window_argument
    @fix_return_object
    def coherence(self, other, fftlength=None, overlap=None, window=None, **kwargs):
        """Calculate the frequency-coherence between this `TimeSeries` and another.

        Parameters
        ----------
        other : `TimeSeries`
            `TimeSeries` signal to calculate coherence with

        fftlength : `float`, optional
            number of seconds in single FFT. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, calculates single FFT covering full duration and then performs averaging in frequency domain.
            * For LPSD averaging method (`method='lpsd'`): user-specified value ignored, algorithm calculates optimal segment lengths.
            * For other averaging methods: defaults to a single FFT covering the full duration (**NOTE**: THIS DEFAULT VALUE IN COHERENCE CALCULATION DOES NOT MAKE SENSE FOR MOST REAL APPLICATIONS!)

        overlap : `float`, optional
            number of seconds of overlap between FFTs. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, no overlap possible because a single FFT is calculated.
            * For other averaging methods: defaults to the recommended overlap for the given window (if given), or 0

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. Behavior depends on averaging method:

            * For LPSD averaging method (`method='lpsd'`): only `str` type is allowed. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
            * For other averaging methods: see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method: `str`, optional
                averaging method for coherence calculation. See above for important difference in arguments.
                Defaults to gwpy.signal.spectral.coherence

            any other keyword arguments accepted by the respective averaging methods.
            See definitions of corresponding method (`method` keyword). If `method` is not specified,
            defaults to gwpy.signal.spectral.coherence

        Returns
        -------
        coherence : FrequencySeries
            the coherence `FrequencySeries` of this `TimeSeries` with the other
        """

        method = kwargs.pop("method", None)

        # calculate coherence
        if method == "daniell" or method == "lpsd":
            coherence = self._coherence(
                other,
                window=window,
                method=method,
                **kwargs,
            )
        elif method is None:
            # using default GWpy method; in that case, default fftlength will may also be used
            # inform the user of dangers
            if fftlength is None:
                warn(
                    "No 'fftlength' specified, note that in this case single FFT covering whole time series is used"
                )

            coherence = spectral.psd(
                (self, other),
                spectral.coherence,
                fftlength=fftlength,
                overlap=overlap,
                window=window,
                **kwargs,
            )
        else:
            raise NotImplementedError(
                "Only 'daniell' and 'lpsd' averaging methods are currently implemented in addition to default"
            )
        return coherence

    def _coherence(
        self,
        other,
        window=None,
        method="daniell",
        **kwargs,
    ):
        """Calculate the frequency-coherence between this `TimeSeries` and another with "custom" averaging methods. This method then calculates coherence using the formula:

        `coherence = np.abs(csd) ** 2 / psd1 / psd2`

        Parameters
        ----------
        other : `TimeSeries`
            `TimeSeries` signal to calculate coherence with

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. Behavior depends on averaging method:

            * For LPSD averaging method (`method='lpsd'`): only `str` type is allowed. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
            * For other averaging methods: see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            any other keyword arguments accepted by the respective averaging methods.
            See definitions of corresponding method (`method` keyword).

        Returns
        -------
        coherence : FrequencySeries
            the coherence `FrequencySeries` of this `TimeSeries` with the other
        """

        if method == "daniell":
            method_func = daniell
        elif method == "lpsd":
            method_func = lpsd
        else:
            raise NotImplementedError(
                "Custom coherence calculation is only implemented for 'daniell' and 'lpsd' "
                "averaging methods."
            )
        if kwargs.get("overlap"):
            raise ValueError(
                "overlap argument is not supported by custom averaging methods; "
                "For Daniell averaging method: no overlap possible because a single FFT is calculated."
                "For LPSD averaging method: defaults to the recommended overlap for the given window (if given), or 0"
            )
        if kwargs.get("fftlength"):
            raise ValueError(
                "fftlength argument is not supported by custom averaging methods; "
                "For Daniell averaging method: single FFT is calculated and then averaging performed in frequency domain."
                "For LPSD averaging method: algorithm calculates optimal segment lengths."
            )
        csd = spectral.psd(
            (self, other),
            method_func=method_func,
            fftlength=None,
            overlap=None,
            window=window,
            **kwargs,
        )
        psd1 = spectral.psd(
            self,
            method_func=method_func,
            fftlength=None,
            overlap=None,
            window=window,
            **kwargs,
        )
        psd2 = spectral.psd(
            other,
            method_func=method_func,
            fftlength=None,
            overlap=None,
            window=window,
            **kwargs,
        )
        coherence = np.abs(csd) ** 2 / psd1 / psd2
        coherence.name = f"Coherence between {self.name} and {other.name}"

        return coherence

    @check_not_implemented_method
    @fix_return_object
    def fft(self, *args, **kwargs):
        """Compute the one-dimensional discrete Fourier transform of this `TimeSeries`.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().fft(*args, **kwargs)

    def fetch(*args, **kwargs):  # pylint: disable=no-self-argument
        """Fetch time series (e.g. from NDS server).
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        #  note this is not a method, but rather just a function,
        #  so have to wrap differently than other methods
        return TimeSeries.from_other(gwpy.timeseries.TimeSeries.fetch(*args, **kwargs))

    @check_not_implemented_method
    @fix_return_object
    def average_fft(self, *args, **kwargs):
        """Compute the averaged one-dimensional DFT of this `TimeSeries`.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().average_fft(*args, **kwargs)

    def chunk(self, stride):
        """Split these time series in several chunks and return a list.

        Parameters
        ----------
        stride: `float`
            stride for spectrogram, in seconds

        Returns
        -------
        list : `list` of `TimeSeries`
            Chunks of TimeSeries in a list
        """
        from gwpy.signal.spectral._ui import seconds_to_samples

        n_stride = seconds_to_samples(stride, self.sample_rate)
        n_samples = len(self.value)
        if n_stride > n_samples:
            raise ValueError("Specified stride is longer than time series length.")

        n_chunks = (
            int(n_samples / n_stride)
            if n_samples % n_stride == 0
            else int(n_samples / n_stride) + 1
        )
        chunks = []
        for i in range(n_chunks):
            chunk = self[i * n_stride : (i + 1) * n_stride]
            chunks.append(chunk)
        return chunks

    @fix_return_object
    def spectrogram(self, stride, **kwargs):
        """Calculate the average power spectrogram of this `TimeSeries`.

        Parameters
        ----------
        stride: `float`
            stride for spectrogram, in seconds

        **kwargs
            method : `str`, optional
                FFT-averaging method (default: ``'median'``). The accepted ``method`` arguments are:

                - ``'bartlett'`` : a mean average of non-overlapping periodograms
                - ``'median'`` : a median average of overlapping periodograms
                - ``'welch'`` : a mean average of overlapping periodograms
                - ``'lpsd'`` :  average of overlapping periodograms binned logarithmically in frequency
                - ``'daniell'`` : calculates single fft for the whole time series and averages in frequency domain

            any other keyword arguments accepted by GWpy TimeSeries.Spectrogram method,
            and the respective averaging methods.
            See definitions of corresponding method (`method` keyword). If `method` is not specified,
            defaults to gwpy.signal.spectral.csd

        Returns
        -------
        spectrogram :  Spectrogram
            average power spectrogram of this `TimeSeries`
        """
        #  enable multiprocessing for spectrogram
        from gwpy.utils import mp
        from .spectrogram import Spectrogram

        if "method" not in kwargs or (
            kwargs["method"] != "lpsd" and kwargs["method"] != "daniell"
        ):
            return super().spectrogram(stride, **kwargs)

        # form list of inputs for each PSD calculation
        chunks = self.chunk(stride)

        # bundle inputs for _psd
        inputs = [(chunk, kwargs) for chunk in chunks]

        # calculate spectrogram
        n_proc = kwargs.pop("n_proc", 1)
        psds = mp.multiprocess_with_queues(n_proc, _psd, inputs)
        epoch = self.t0.value
        return Spectrogram.from_spectra(*psds, epoch=epoch, dt=stride)

    @check_not_implemented_method
    @fix_return_object
    def spectrogram2(self, *args, **kwargs):
        """Calculate the non-averaged power `Spectrogram` of this `TimeSeries`
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().spectrogram2(*args, **kwargs)

    @check_not_implemented_method
    @fix_return_object
    def fftgram(self, *args, **kwargs):
        """Calculate the Fourier-gram of this `TimeSeries`.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().fftgram(*args, **kwargs)

    @check_not_implemented_method
    @fix_return_object
    def spectral_variance(self, *args, **kwargs):
        """Calculate the `SpectralVariance` of this `TimeSeries`.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().spectral_variance(*args, **kwargs)

    @check_not_implemented_method
    @fix_return_object
    def rayleigh_spectrum(self, *args, **kwargs):
        """Calculate the Rayleigh `FrequencySeries` for this `TimeSeries`.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().rayleigh_spectrum(*args, **kwargs)

    @check_not_implemented_method
    @fix_return_object
    def rayleigh_spectrogram(self, *args, **kwargs):
        """Calculate the Rayleigh statistic spectrogram of this `TimeSeries`.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().rayleigh_spectrogram(*args, **kwargs)

    @check_not_implemented_method
    @fix_return_object
    def csd_spectrogram(self, *args, **kwargs):
        """Calculate the cross spectral density spectrogram of this`TimeSeries` with 'other'.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().csd_spectrogram(*args, **kwargs)

    @fix_window_argument
    @fix_return_object
    def transfer_function(
        self, other, fftlength=None, overlap=None, window="hann", **kwargs
    ):
        """Calculate the transfer function between this `TimeSeries` and
        another.

        This `TimeSeries` is the 'A-channel', serving as the reference
        (denominator) while the other time series is the test (numerator)

        Parameters
        ----------
        other : `TimeSeries`
            `TimeSeries` signal to calculate the transfer function with

        fftlength : `float`, optional
            number of seconds in single FFT. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, calculates single FFT covering full duration and then performs averaging in frequency domain.
            * For LPSD averaging method (`method='lpsd'`): user-specified value ignored, algorithm calculates optimal segment lengths.
            * For other averaging methods: defaults to a single FFT covering the full duration

        overlap : `float`, optional
            number of seconds of overlap between FFTs. Default behavior:

            * For Daniell averaging method (`method='daniell'`): user-specified value ignored, no overlap possible because a single FFT is calculated.
            * For other averaging methods: defaults to the recommended overlap for the given window (if given), or 0

        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. Behavior depends on averaging method:

            * For LPSD averaging method (`method='lpsd'`): only `str` type is allowed. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
            * For other averaging methods: see :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.

        **kwargs
            method : `str`, optional
                FFT-averaging method (default: ``'mean'``). The accepted ``method`` arguments are:

                - ``'bartlett'`` : a mean average of non-overlapping periodograms
                - ``'median'`` : a median average of overlapping periodograms
                - ``'welch'`` : a mean average of overlapping periodograms
                - ``'lpsd'`` :  average of overlapping periodograms binned logarithmically in frequency
                - ``'daniell'`` : calculates single fft for the whole time series and averages in frequency domain

            average : `str`, optional
                FFT-averaging method for scipy (default: ``'mean'``) passed to underlying csd() and psd() methods,
                unless "daniell" or "lpsd" is used.

        Returns
        -------
        transfer_function : `FrequencySeries`
            the transfer function `FrequencySeries` of this `TimeSeries`
            with the other

        Notes
        -----
        If `self` and `other` have difference
        :attr:`TimeSeries.sample_rate` values, the higher sampled
        `TimeSeries` will be down-sampled to match the lower.
        """
        average = kwargs.pop(
            "average", "mean"
        )  # apply gwpy default ("mean") if none specified
        return super().transfer_function(
            other,
            fftlength=fftlength,
            overlap=overlap,
            window=window,
            average=average,
            **kwargs,
        )

    @check_not_implemented_method
    @fix_return_object
    def coherence_spectrogram(self, *args, **kwargs):
        """Calculate the coherence spectrogram between this `TimeSeries` and other.
        Simply wrapping GWpy method, see GWpy docs for details.
        """
        return super().coherence_spectrogram(*args, **kwargs)



"""
Wrapper for LPSD method, see https://gitlab.com/uhh-gwd/lpsd

| Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""

from warnings import warn

import inspect
import pandas as pd
import numpy as np

from lpsd import lcsd
from lpsd._lcsd import LCSD
from gwpy.signal.spectral._utils import scale_timeseries_unit

from spicypy.signal.frequency_series import FrequencySeries


def lpsd(*args, **kwargs):  # pylint: disable=too-many-branches
    """Calculate the CSD of a `TimeSeries` using LPSD method.

    Parameters
    ----------
    args : `list`
        timeseries: `TimeSeries`
            input time series (if only one specified, calculates PSD)
        other: `TimeSeries`, optional
            second input time series
        nfft: `int`
            number of samples per FFT, user-specified value ignored; calculated by the algorithm instead

    kwargs : dict
        fftlength : `float`, optional
            number of seconds in single FFT. User-specified value ignored, algorithm calculates optimal segment lengths.
        overlap : `float`, optional
            number of seconds of overlap between FFTs.
        window : `str`, optional
            Window function to apply to timeseries prior to FFT. Possible values: 'hann', 'hanning', 'ham', 'hamming', 'bartlett', 'blackman', 'kaiser'. Defaults to 'kaiser'.
        additional arguments are passed to :class:`lpsd.lcsd`

    Returns
    -------
    fs: FrequencySeries
        resulting CSD
    """

    try:
        timeseries, nfft = args
        other = timeseries
    except ValueError:
        timeseries, other, nfft = args

    if len(timeseries.value) != len(other.value):
        raise ValueError("Time series must have the same length (number of samples)!")

    # No fftlength for LPSD method.
    if kwargs.get("fftlength") or nfft != len(timeseries.value):
        warn(
            "fftlength/nfft arguments are not supported by LPSD averaging method; "
            "segment lengths are calculated by the algorithm."
        )

    # convert overlap given in number of seconds to percentage
    overlap = kwargs.pop("overlap", 0)
    if overlap > 0:
        total_duration = timeseries.duration
        if overlap > total_duration:
            raise ValueError(
                "Specified overlap (in seconds) exceeds total time series duration!"
            )
        overlap = overlap / total_duration

    # convert window to numpy function
    window = kwargs.pop("window_", None)
    window = "kaiser" if window is None else window
    # clean up default value from kwargs
    if "window" in kwargs:
        kwargs.pop("window")
    if not isinstance(window, str):
        warn(
            "Specifying window as an array for Daniell averaging method is not supported, defaulting to 'hann' window"
        )
        window = "kaiser"

    window_to_func = {
        "kaiser": np.kaiser,
        "hann": np.hanning,
        "hanning": np.hanning,
        "hamm": np.hamming,
        "hamming": np.hamming,
        "bartlett": np.bartlett,
        "blackman": np.blackman,
    }
    try:
        window_function = window_to_func[window]
    except KeyError as exc:
        raise KeyError(
            "Window " + window + "is not supported for LPSD averaging method"
        ) from exc

    # Convert inputs to pandas.DataFrame
    df = pd.DataFrame()
    df["x1"] = timeseries.value
    df["x2"] = other.value
    df.index = timeseries.times.value

    # clean up kwargs: get ones that are allowed for LPSD
    allowed_kwargs = inspect.getfullargspec(LCSD.__init__).args
    lpsd_kwargs = kwargs.copy()
    for k in kwargs:
        if k not in allowed_kwargs:
            lpsd_kwargs.pop(k)
    csd = lcsd(df, overlap=overlap, window_function=window_function, **lpsd_kwargs)

    # generate FrequencySeries and return
    unit = scale_timeseries_unit(
        timeseries.unit,
        kwargs.pop("scaling", "density"),
    )
    fs = FrequencySeries(
        csd["psd"].values,
        unit=unit,
        frequencies=csd.index.values,
        name=timeseries.name,
        epoch=timeseries.epoch,
        channel=timeseries.channel,
    )
    return fs



"""
| Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""

import logging as log
import sys
from importlib import metadata

from spicypy import control, signal

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    __version__ = "unknown"
    log.warning("Version not known, importlib.metadata is not working correctly.")

__all__ = ["control", "signal"]



"""
Class creating Wiener filter from reference time series to signal time series.

Wiener filter gives an estimate of target channel from one or more references, based on linear coherence.
This estimate is optimal for linear and stationary signals. The most common application is to have reference(s) for
noise which enables then subtraction of this noise from target channel. More generally, Wiener filter generates
a time-domain filter that is equivalent to frequency-domain transfer function between two channels, or several such
transfer functions if more than one reference is used (one per reference).

The most common Wiener filter algorithm available (e.g. in image processing etc) relies on one reference, or
even no reference if noise is being estimated directly in the target channel. In other words it is a single-input,
single-output Wiener filter, or sometimes the output has higher dimensionality, such as the case with filtering
images (2D). MISO filter (based on algorithm in https://dcc.ligo.org/LIGO-T070192/public), where multiple references
can be supplied to estimate target channel. See also `J. Harms, Terrestrial gravity fluctuations, Living Reviews in
Relativity 22, 6 (2019)`.

This complicates the calculation as the matrix to invert is no longer a simple Toeplitz matrix,
but a block-Toeplitz matrix. This structure however is still exploited to speed up calculation significantly compared
to general matrix inversion. Implementation is based on iterative solvers from `scipy.sparse.linalg` that don't require
full matrix knowledge for every iteration and `scipy.sparse.linalg.LinearOperator` to generate matrix elements based on
a rule (operator), which permits exploiting symmetries of the matrix. In case of longer time series than necessary for
filter generation, several filters will be generated and the best automatically selected. Application of the filter
to references is also optimized using `numpy` arrays and optionally can be done with `multiprocessing`.

Authors:
    | Artem Basalaev <artem[dot]basalaev[at]pm.me>
    | Tim J. Kuhlbusch <tim[dot]kuhlbusch[at]rwth-aachen.de>
"""
from copy import copy
from typing import Iterable
import multiprocessing
from dataclasses import dataclass, asdict

import numpy as np

# import LinearOperator and all sorts of different solvers
from scipy.sparse.linalg import (
    LinearOperator,
    gmres,
    cgs,
    cg,
    gcrotmk,
    tfqmr,
    bicgstab,
    lgmres,
    minres,
)
from scipy.signal import correlate, correlation_lags
from scipy.linalg import matmul_toeplitz
from scipy import optimize
from spicypy.signal.time_series import TimeSeries
from spicypy.signal import Filter

_solver_list = {
    "gmres": gmres,
    "cgs": cgs,
    "cg": cg,
    "gcrotmk": gcrotmk,
    "tfqmr": tfqmr,
    "bicgstab": bicgstab,
    "lgmres": lgmres,
    "minres": minres,
}


@dataclass
class WienerFilter(Filter):  # pylint: disable=too-many-instance-attributes
    """
    Class creating Wiener filter from reference time series to signal time series.

        Parameters
        ----------
        test: `TimeSeries`
            input time series
        reference: `list`, `TimeSeriesDict`
            reference time series
        kwargs : `dict`
            n_taps : `int`
                Number of taps for this Wiener filter, should be less than time series length.
            zero_padding : `bool`, optional
                Whether to use zero padding; if set to true, resulting time series will begin at the same time as input time series, but will be filled with zeros for n_taps (default: true).
            use_multiprocessing : `bool`, optional
                Whether to use multiprocessing to apply filters (default:false).
            n_proc: `int`, optional
                Number of parallel processes to use. Defaults to number of CPUs.
            subtract_mean: `bool`, optional
                Subtract mean from data for internal processing, to ensure that floating point numbers are not too large. For the output time series operation is inverted, therefore besides improving numerical stability it is not affecting the results. Defaults to true.
            normalize_data: `bool`, optional
                Divide data by its standard deviation for internal processing, to ensure that floating point numbers are not too large. For the output time series operation is inverted, therefore besides improving numerical stability it is not affecting the results. Defaults to true.
            use_norm_factor: `bool`, optional
                Renormalize resulting Wiener filter by finding the best-fitting normalization coefficient for ASD on training data. May result in improvement in case of MISO filter with noisy references, otherwise not recommended. Defaults to false.
            verbose: `bool`, optional
                Enable verbose output (default: false).
            solver: `str`, optional
                Solver to use for Wiener-Hopf equations. Default: `tfqmr`. Other available: `gmres`, `cgs`, `cg`, `gcrotmk`, `tfqmr`, `bicgstab`, `lgmres`, `minres`.
    """

    # the values defined here will be exported in the JSON dump
    # through the asdict() method and this being a dataclass
    W: dict
    best_filter_index: int
    kwargs: dict
    norm_factor: dict
    normalize: bool
    use_norm_factor: bool
    zero_padding: bool
    solver = None
    solver_arg: str
    verbose: bool

    filter_name: str = "WienerFilter"
    additional_json_keys: Iterable[str] = (
        "W",
        "best_filter_index",
        "kwargs",
        "norm_factor",
        "normalize",
        "use_norm_factor",
        "zero_padding",
        "solver",
        "solver_arg",
        "verbose",
        "filter_name",
    )

    def __init__(self, test, reference, **kwargs):
        """Init method for Wiener filter"""
        super().__init__(test, reference, **kwargs)

        self.use_multiprocessing = kwargs.pop("use_multiprocessing", False)
        self.verbose = kwargs.pop("verbose", False)

        if self.loaded_from_file:
            # NOTE: this might not be the best way to handle the whole keys become strings
            # through the json dump thing..
            self.W = {int(k): np.array(v) for k, v in self.W.items()}
            self.norm_factor = {int(k): v for k, v in self.norm_factor.items()}
            self.best_filter_index = int(
                self.best_filter_index
            )  # old save files have this index as a string value
            self.initialized = True
            return

        self.kwargs = kwargs
        self.test_array = None
        self.reference_array = None

        self.unique_rmi = {}
        self.P = {}
        self.labels = {}
        self.W = {}
        self.norm_factor = {}
        self.test_std = 1.0
        self.test_mean = 0.0
        self.refs_std = []
        self.refs_mean = []
        self.best_filter_index = 0

        self._check_inputs()
        # Note for following attributes: they are needed only to be able to use pool.map(..) function and
        # pass information which normally would be done with function arguments. Maybe there's a nicer way to do it.
        self.parallel_idx = None
        self.parallel_slice_length = None
        self.parallel_filter_index = None
        self.parallel_refs = None

        self.initialized = False  # to indicate whether create_filters() was called

    def _check_inputs(self):
        """Parse input arguments for this Wiener filter."""

        if self.loaded_from_file:  # Do nothing; no inputs to check if loaded from file
            return

        # make a list of all inputs and compare values to test
        inputs_list = self.reference + [self.test]
        self._check_timeseries_consistency(inputs_list)
        self.sample_rate = self.test.sample_rate.value
        self.n_references = len(self.reference)
        self.unit = self.test.unit.to_string()

        # get parameters for Wiener filter calculation
        self.zero_padding = self.kwargs.pop("zero_padding", True)
        self.n_proc = self.kwargs.pop("n_proc", multiprocessing.cpu_count())
        if self.n_proc == 1:
            self.use_multiprocessing = False
        self.subtract_mean = self.kwargs.pop("subtract_mean", True)
        self.normalize = self.kwargs.pop("normalize_data", True)
        self.use_norm_factor = self.kwargs.pop("use_norm_factor", False)

        self.solver = tfqmr
        self.solver_arg = self.kwargs.pop("solver", "tfqmr")
        if self.solver_arg in _solver_list:
            self.solver = _solver_list[self.solver_arg]
        else:
            raise NotImplementedError(
                "Requested solver " + self.solver_arg + " is not supported"
            )
        if self.use_norm_factor and self.normalize:
            raise NotImplementedError(
                "Additional norm factor can only be derived"
                " when Wiener filter is applied to already normalized data"
            )

    def save(self, filename: str):
        """Save the filter as a JSON text file.

        Parameters
        ----------
        filename:
            Target filename. If file name does not end in `.json`, it will be appended.
        """
        if not self.initialized:
            raise RuntimeError(
                "save() can only be called on an instance that has been initialized by executing create_filters()!"
            )
        super().save(filename)

    def create_filters(self):
        """Calculate Wiener filter(s) based on inputs."""
        self.initialized = True

        if self.loaded_from_file:
            raise NotImplementedError(
                "Cannot generate filters because this Wiener filter was loaded from file;"
                " filters are already generated earlier. Try calling `.apply(..)` instead."
            )
        self._prepare_data()
        # each filter requires n_taps + 1 time steps to create
        if len(self.test) > (
            2 * self.n_taps
        ):  # will create multiple filters and choose the best one
            # first calculate how many times n_taps
            n_filters = int(np.floor(len(self.test) / self.n_taps))
            # then check if there's one more time step for last filter, if not make one less
            if len(self.test) <= n_filters * self.n_taps:
                n_filters -= 1
            print("Creating Wiener Filters...")
            for i in range(n_filters):
                self._create_filter(
                    i,
                    self.test_array,
                    self.reference_array,
                    i * self.n_taps,
                    (i + 1) * self.n_taps,
                )
            self.determine_best_filter()
        else:  # create one filter
            self._create_filter(
                0, self.test_array, self.reference_array, 0, self.n_taps
            )

    def as_dict(self) -> dict:
        """Return a JSON compatible dict representation of the filter."""
        filter_dict = asdict(self)
        filter_dict["W"] = {k: v.tolist() for k, v in filter_dict["W"].items()}
        return filter_dict

    def determine_best_filter(self):
        """Find the best-performing Wiener Filter on input data."""

        if self.loaded_from_file:
            raise NotImplementedError(
                "Cannot determine the best filter because this Wiener filter was "
                "loaded from file; best filters are already found earlier. "
                "Try calling `.apply(..)` instead."
            )

        print("Determining the best filter...")
        mse_arr = np.zeros(len(self.W))
        test_asd = self.test.asd(method="lpsd")
        for idx in self.W:
            result_ts = self.apply(
                inputs_list=None,
                index=idx,
                zero_padding=True,
                normalized_output=True,
                to_training_data=True,
            )
            result_asd = result_ts.asd(method="lpsd")
            mse = 0.0
            if self.use_norm_factor:

                def MSE_to_minimize(norm_factor):
                    return MSE(
                        test_asd.value,
                        norm_factor
                        * result_asd.value,  # pylint: disable=cell-var-from-loop
                    )

                norm = optimize.minimize_scalar(MSE_to_minimize)
                self.norm_factor[idx] = norm.x
                mse = MSE_to_minimize(norm.x)
                if self.verbose:
                    print(
                        idx,
                        "MSE:",
                        MSE_to_minimize(norm.x),
                        "normalization factor:",
                        norm.x,
                    )
            else:
                mse = MSE(test_asd.value, result_asd.value)
                if self.verbose:
                    print(idx, "MSE:", mse)
            mse_arr[idx] = mse
        self.best_filter_index = np.argmin(mse_arr)
        print("Done. Best filter index:", self.best_filter_index)

    def _create_filter(
        self, index, tot_signal, tot_refs, start, end
    ):  # pylint: disable=too-many-positional-arguments
        """Calculate individual Wiener filter

        Parameters
        ----------
        index: `int`
            Current filter index.
        tot_signal: `np.array` of `float`
            Target time series, which filter should recreate.
        tot_refs: `np.array` of `float`
            One or multiple reference time series.
        start: `int`
            First index in the array for this filter.
        end: `int`
            Last index in the array for this filter.
        """

        if self.verbose:
            print("Creating Wiener filter: ", index)
        signal = tot_signal[start : end + 1]
        refs = tot_refs[start : end + 1]
        n_refs = refs.shape[1]

        self._wiener_components(refs, signal, index)
        unique_rmi = self.unique_rmi[index]
        unique_colrows = unique_rmi.shape[0]
        labels = self.labels[index]

        def R_block_times_vec(q, r, vec):
            colrow_idx = None
            transpose = False
            for i in range(unique_colrows):
                if labels[i][0] == q and labels[i][1] == r:
                    colrow_idx = i
                    break
                if q != r and labels[i][0] == r and labels[i][1] == q:
                    colrow_idx = i
                    transpose = True
                    break
            if colrow_idx is None:
                raise ValueError("Block ", q, r, "is out of bounds!")
            rmi_left = unique_rmi[colrow_idx][0 : self.n_taps + 1]
            rmi_right = unique_rmi[colrow_idx][self.n_taps :]
            if transpose:
                return matmul_toeplitz((rmi_right, np.flip(rmi_left)), vec)
            return matmul_toeplitz((np.flip(rmi_left), rmi_right), vec)

        def R_times_vec(vec):
            vec = np.ndarray.flatten(vec)
            outvec = np.zeros(shape=vec.shape[0])
            for i in range(n_refs):
                for j in range(n_refs):
                    outvec[
                        i * (self.n_taps + 1) : (i + 1) * (self.n_taps + 1)
                    ] = outvec[
                        i * (self.n_taps + 1) : (i + 1) * (self.n_taps + 1)
                    ] + R_block_times_vec(
                        i, j, vec[j * (self.n_taps + 1) : (j + 1) * (self.n_taps + 1)]
                    )
            return outvec

        operator_dim = (self.n_taps + 1) * refs.shape[1]
        Rv = LinearOperator(shape=(operator_dim, operator_dim), matvec=R_times_vec)
        if self.verbose:
            print("inverting the matrix")
        W = self.solver(Rv, self.P[index])
        self.W[index] = W[0]

    def _wiener_components(self, refs, signal, index):
        """Prepare components for the Wiener-Hopf equations: covariance matrix, cross-corr vector.

        Parameters
        ----------
        refs: `np.array` of `float`
            One or multiple reference time series.
        signal: `np.array` of `float`
            Target time series, which filter should recreate.
        index: `int`
            Current filter index.
        """

        n_refs = refs.shape[1]
        n_time_steps = refs.shape[0]

        # unique components ("rmi") rows/cols of covariance matrix
        unique_colrows = n_refs * (n_refs + 1) // 2
        colrow_length = 2 * (self.n_taps + 1) - 1
        unique_rmi = np.zeros(shape=(unique_colrows, colrow_length))

        # input covariance matrix
        lags = correlation_lags(n_time_steps, n_time_steps)
        max_lag = np.where(np.abs(lags) <= self.n_taps)
        k = 0
        for m in range(n_refs):
            for i in range(m, n_refs):
                if self.verbose:
                    print("calculating R" + str(m) + str(i))
                rmi = correlate(refs[:, m], refs[:, i])
                rmi = np.ndarray.flatten(np.take(rmi, max_lag))
                unique_rmi[k, :] = rmi
                k = k + 1

        # cross−correlation vector
        P = np.zeros(n_refs * (self.n_taps + 1))
        if self.verbose:
            print("calculating cross-corr vector")
        for i in range(n_refs):
            top = i * (self.n_taps + 1)
            bottom = (i + 1) * (self.n_taps + 1)
            p = correlate(signal, refs[:, i])
            p = np.ndarray.flatten(np.take(p, max_lag))
            P[top:bottom] = np.conjugate(p[self.n_taps :])

        k = 0
        labels = np.zeros(shape=(unique_colrows, 2), dtype=np.int8)
        for m in range(n_refs):
            for i in range(m, n_refs):
                labels[k] = [m, i]
                k = k + 1

        self.unique_rmi[index] = unique_rmi
        self.P[index] = P
        self.labels[index] = labels

    def _apply_parallel(self, i):
        """Apply Wiener filter to partial data by one parallel process.
        Can be called once to apply to all data by a single process.

        Parameters
        ----------
        i: `int`
            index of the parallel process (default: 0 for single process)
        """

        n_time_steps = self.parallel_refs.shape[0]
        n_refs = self.parallel_refs.shape[1]
        W = self.W[self.parallel_filter_index]
        indices = self.parallel_idx[
            self.parallel_slice_length * i : self.parallel_slice_length * (i + 1)
        ]
        result_timeseries = np.zeros(n_time_steps)

        for n in indices:  # time steps for which we can calculate
            # output signal
            if self.verbose and n % 50000 == 0:
                print(
                    "Applying filter " + str(self.parallel_filter_index) + ": step",
                    n - self.n_taps,
                    "out of",
                    n_time_steps - self.n_taps,
                )
            for m in range(n_refs):  # number of ref channels
                result_timeseries[n] = result_timeseries[n] + W[
                    (self.n_taps + 1) * m : (self.n_taps + 1) * (m + 1)
                ].dot(np.flip(self.parallel_refs[n - self.n_taps : n + 1, m]))
        return indices, result_timeseries

    def _post_process_data(  # pylint: disable=too-many-positional-arguments
        self,
        result_timeseries,
        zero_padding,
        to_training_data,
        normalized_output,
        index,
        inputs_list,
    ):
        """Post-process data (revert normalization, package as `TimeSeries`, etc.).

        Parameters
        ----------
        result_timeseries: `np.array` of `float`
            Raw time series - result of the applied filter
        zero_padding: `bool`
            Whether to add zeros in front (results in time series of same length as input).
        to_training_data: `bool`
            Whether the filter was applied to training data.
        normalized_output: `bool`
            Whether the data were normalized before creating filter - in that case inverse should be applied to the result.
        index: `int`
            Index of the applied filter, in case it has individual norm factor to be applied.
        inputs_list: `np.array` of `float`
            Array of inputs (references) to which this filter is applied.
        """

        if not zero_padding:
            result_timeseries = result_timeseries[self.n_taps :]

        unit = self.unit
        if to_training_data:
            t0 = copy(self.test.t0)
            sample_rate = self.test.sample_rate
        else:
            t0 = copy(inputs_list[0].t0)
            sample_rate = inputs_list[0].sample_rate

        if not zero_padding:  # advance start time by filter length
            t0 += self.n_taps / sample_rate

        if not normalized_output:
            norm_factor = 1.0
            if index in self.norm_factor:
                norm_factor = self.norm_factor[index]
                if self.verbose:
                    print("Multiplying by norm factor=", norm_factor)
            result_timeseries *= norm_factor
            if self.normalize:
                result_timeseries = result_timeseries * self.test_std
                if self.verbose:
                    print("Multiplying by test std=", self.test_std)
            if self.subtract_mean:
                result_timeseries += self.test_mean
                if self.verbose:
                    print("Adding mean=", self.test_mean)

        return TimeSeries(
            result_timeseries,
            sample_rate=sample_rate,
            unit=unit,
            t0=t0,
            name="Wiener filter output",
        )

    @Filter._check_apply_parameters
    def apply(  # pylint: disable=too-many-positional-arguments
        self,
        inputs_list=None,
        zero_padding=None,
        normalized_output=False,
        to_training_data=False,
        use_multiprocessing=None,
        index=None,
    ):
        """Apply Wiener filter to data.

        Parameters
        ----------
        inputs_list: `list` of `TimeSeries`, optional
            time series to apply Wiener filter to, not needed if applying to training data
        index: `int`, optional
            use specific Wiener filter by index, by default using the best filter determined earlier
        zero_padding: `bool`, optional
            if true, return time series of the same length as inputs,
            with zeros in the beginning (Wiener filter needs to "accumulate" data equal to its n_taps first);
            if false, returned time series will be shorter and with shifted t0 by n_taps/sample_rate
        normalized_output: `bool`, optional
            if true, do not add mean and multiply by standard deviation, as if the data were already normalized
        to_training_data: `bool`, optional
            apply filters to training data - runs after filter creation to de3termine best filter,
            but can also be called explicitly
        use_multiprocessing: `bool`, optional
            use multiprocessing (with number of parallel processes set with `self.n_proc`)

        Returns
        -------
        prediction :  `TimeSeries`
            Wiener filter output
        """
        if not self.initialized:
            raise RuntimeError(
                "apply() can only be called on an instance that has been initialized by executing create_filters()!"
            )

        if use_multiprocessing is None:
            use_multiprocessing = self.use_multiprocessing

        if index is None:  # apply the best filter
            index = self.best_filter_index

        refs = self._preprocess_data(to_training_data, inputs_list)

        n_time_steps = refs.shape[0]
        idx = np.arange(
            self.n_taps, n_time_steps
        )  # time steps for which we can calculate output signal
        self.parallel_idx = idx
        self.parallel_filter_index = index
        self.parallel_refs = refs

        if use_multiprocessing and len(idx) < self.n_proc:
            print(
                "Requested multiprocessing but time series is not long enough to benefit from it. "
                "You can try to set n_proc to lower values to run on fewer processes than "
                + str(self.n_proc)
                + ". Running in single process mode."
            )
        if use_multiprocessing and len(idx) > self.n_proc:
            print(
                "Using multiprocessing. Your computer may become unresponsive due to load, this is expected. Please "
                "wait until the operation is complete."
            )
            self.parallel_slice_length = int(np.ceil(len(idx) / float(self.n_proc)))
            n_slices = int(np.ceil(len(idx) / float(self.parallel_slice_length)))
            result_timeseries = np.zeros(self.parallel_refs.shape[0])

            with multiprocessing.Pool(processes=self.n_proc) as pool:
                result_list = pool.map(self._apply_parallel, range(n_slices))
            for indices, ts in result_list:
                result_timeseries[indices[0] : indices[-1]] = ts[
                    indices[0] : indices[-1]
                ]

        else:
            self.parallel_slice_length = len(idx)
            _, result_timeseries = self._apply_parallel(0)

        if zero_padding is None:
            zero_padding = self.zero_padding

        return self._post_process_data(
            result_timeseries,
            zero_padding,
            to_training_data,
            normalized_output,
            index,
            inputs_list,
        )


def MSE(sig_asd, pred_asd):
    """Calculate mean squared error on FrequencySeries

    Parameters
    ----------
    sig_asd: `FrequencySeries`
        "signal" frequency series
    pred_asd: `FrequencySeries`
        "prediction" frequency series

    Returns
    -------
    mse :  float
        mean squared error
    """
    return np.mean(np.square(sig_asd - pred_asd))



"""
Class extending functionality of :obj:`gwpy.frequencyseries.frequencyseries.FrequencySeries` from GWpy

Authors:
    | Artem Basalaev <artem[dot]basalaev[at]pm.me>
    | Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
    | Shreevathsa Chalathadka Subrahmanya <schalath[at]physnet.uni-hamburg.de>
"""
import copy
from warnings import warn
import gwpy.frequencyseries
import numpy as np
from astropy.units import UnitsError
from astropy import units as u


class FrequencySeries(gwpy.frequencyseries.FrequencySeries):
    """
    Class to model spectra of signals (frequency series), inherits from GWpy

    """

    @classmethod
    def from_other(cls, fs):
        """Create FrequencySeries from another FrequencySeries, in particular gwpy.FrequencySeries

        Parameters
        ----------
        fs : FrequencySeries, gwpy.frequencyseries.FrequencySeries
            Other FrequencySeries object
        Returns
        -------
        FrequencySeries
            Copy Spicypy FrequencySeries object
        """
        fs_spicypy = cls(fs.value)
        fs_spicypy.__dict__ = copy.deepcopy(fs.__dict__)
        return fs_spicypy

    def times_iomega(self, omega_exponent=1):
        """Differentiate or integrate spectrum wrt. time

        Parameters
        ----------
        omega_exponent : int
            Power of i*omega in multiplication factor, i.e. spectrum is multiplied by (i*omega)^omega_exponent.

            * values > 0 correspond to differentiation omega_exponent times,
            * values < 0 correspond to integration omega_exponent times.

        Returns
        -------
        FrequencySeries
            Integrate/differentiate frequency series.
        """
        if omega_exponent == 0:
            warn("Specified omega_exponent is 0. Returning as it is")
            return self
        omega = self.frequencies * 2 * np.pi
        dc_bin = False
        if omega[0] == 0.0:  # avoid potential divide by zero
            omega = omega[1:]
            dc_bin = True

        # let's first check if any of the values are complex
        if np.iscomplex(self.value).any():
            omega = (
                1j * omega
            )  # yes, there are complex values -  do it "properly" with i*omega
            # otherwise, ignore the i and apply real omega factor

        # calculate omega factor
        omega_pow = np.power(omega, omega_exponent)

        if dc_bin:
            omega_pow = np.insert(omega_pow, 0, 0.0)
        output = self * omega_pow
        # astropy mixes Hz and s. Manual correction is done below.
        for _ in range(abs(omega_exponent)):
            if omega_exponent > 0:
                output.override_unit(
                    output.unit / u.Hz / u.s  # pylint: disable=no-member
                )  # pylint: disable=no-member
            else:
                output.override_unit(
                    output.unit * u.Hz * u.s  # pylint: disable=no-member
                )  # pylint: disable=no-member
        return output

    def to_displacement(self):
        """Integrate the spectrum once or twice to convert to displacement.
        Input spectrum must have velocity or acceleration units.

        Parameters
        ----------

        Returns
        -------
        FrequencySeries
            Displacement frequency series.
        """

        # first check units - only velocity and acceleration can be converted to displacement
        original_units = ["m*s(-1)/Hz(1/2)", "m*s(-2)/Hz(1/2)"]
        omega_exponents = [
            -1,
            -2,
        ]  # possible exponents to use, depending on whether it is velocity or acceleration
        omega_exponent = None
        for i in range(len(omega_exponents)):
            try:
                self.unit.to(original_units[i])  # pylint: disable=no-member
            except UnitsError:
                continue  # could not convert to this unit, try with another one
            omega_exponent = omega_exponents[
                i
            ]  # converted to this unit - save omega_factor and stop the loop
            break

        if omega_exponent is None:
            raise ValueError(
                "Units of this FrequencySeries ",
                self.unit,
                " cannot be converted to displacement",
            )
        return self.times_iomega(omega_exponent)

    def to_velocity(self):
        """Integrate the spectrum once or differentiate once to convert to velocity.
        Input spectrum must have displacement or acceleration units.

        Parameters
        ----------

        Returns
        -------
        FrequencySeries
            Velocity frequency series.
        """

        # first check units - only displacement and acceleration can be converted to velocity
        original_units = ["m/Hz(1/2)", "m*s(-2)/Hz(1/2)"]
        omega_exponents = [
            1,
            -1,
        ]  # possible exponents to use, depending on whether it is displacement or acceleration
        omega_exponent = None
        for i in range(len(omega_exponents)):
            try:
                self.unit.to(original_units[i])  # pylint: disable=no-member
            except UnitsError:
                continue  # could not convert to this unit, try with another one
            omega_exponent = omega_exponents[
                i
            ]  # converted to this unit - save omega_factor and stop the loop
            break

        if omega_exponent is None:
            raise ValueError(
                "Units of this FrequencySeries ",
                self.unit,
                " cannot be converted to velocity",
            )
        return self.times_iomega(omega_exponent)

    def to_acceleration(self):
        """Differentiate the spectrum once or twice once to convert to acceleration.
        Input spectrum must have displacement or velocity units.

        Parameters
        ----------

        Returns
        -------
        FrequencySeries
            Acceleration frequency series.
        """

        # first check units - only displacement and velocity can be converted to acceleration
        original_units = ["m/Hz(1/2)", "m*s(-1)/Hz(1/2)"]
        omega_exponents = [
            2,
            1,
        ]  # possible exponents to use, depending on whether it is displacement or velocity
        omega_exponent = None
        for i in range(len(omega_exponents)):
            try:
                self.unit.to(original_units[i])  # pylint: disable=no-member
            except UnitsError:
                continue  # could not convert to this unit, try with another one
            omega_exponent = omega_exponents[
                i
            ]  # converted to this unit - save omega_factor and stop the loop
            break

        if omega_exponent is None:
            raise ValueError(
                "Units of this FrequencySeries ",
                self.unit,
                " cannot be converted to acceleration",
            )
        return self.times_iomega(omega_exponent)

    @property
    def is_asd(self):
        """Checks whether the input is ASD or not"""
        if self.unit == "Hz(1/2)":  # only possible for frequency ASD
            return True
        if self.unit == "1/Hz(1/2)":  # only possible for strain (unitless) ASD
            return True
        # let's assume it _is_ ASD, then if we do the following the frequency unit should cancel out
        test_unit = self.unit * u.Unit("Hz (1/2)")
        # now find out which units it had:
        bases = test_unit.decompose().bases
        # now try to force decompose this unit in base units from above _plus_ a frequency unit
        bases.append(u.Hz)  # pylint: disable=no-member
        if "Hz" in test_unit.decompose(bases=bases).bases:
            # as residual frquency unit is still present, not ASD
            return False
        # can be only ASD at this point
        return True

    @property
    def is_psd(self):
        """Checks whether the input is PSD or not"""
        if self.unit == "Hz":  # only possible for frequency PSD
            return True
        if self.unit == "1/Hz":  # only possible for strain (unitless) PSD
            return True
        test_unit = self.unit * u.Unit("Hz (1/2)")
        bases = test_unit.decompose().bases
        if len(bases) == 1 and "s" in bases:
            pass  # bypass for frequency ASD
        else:
            bases.append(u.Hz)  # pylint: disable=no-member
        if "Hz" in test_unit.decompose(bases=bases).bases:
            # as residual frquency unit is still present
            return True
        return False

    def rms(self, start_frequency=None):
        """Calculates the high-to-low RMS of the input spectrum.

        Parameters
        ----------
        start_frequency: float
            frequency to begin accumulating RMS (optional)

        Return
        ------
        rms: FrequencySeries
            High-to-low frequency cumulative RMS
        """
        rms = self.copy()

        # sanity check
        if (self.is_asd is False and self.is_psd is False) or (
            self.is_asd is True and self.is_psd is True
        ):
            raise ValueError(
                "Could not determine from units whether this FrequencySeries is ASD or PSD, cannot calculate RMS!"
            )
        # at this point we know whether it's PSD or ASD (and it can be nothing else)
        # if PSD, first take square root and then proceed with calculation for ASD
        if self.is_psd:
            rms = np.sqrt(rms)
        if start_frequency is not None:
            n_cut = np.where(self.frequencies.value > start_frequency)[0][0]
            rms = self[0:n_cut]

        bin_widths = np.diff(rms.frequencies)
        bin_widths = np.append(bin_widths[0], bin_widths)
        rms = np.sqrt(np.flip(np.cumsum(np.flip(rms * rms * bin_widths))))

        return rms


class SpectralVariance(gwpy.frequencyseries.SpectralVariance):
    """
    Class to model spectral variance (frequency series), copy from GWpy with fix for LPSD averaging method.

    """

    @classmethod
    def from_spectrogram(
        cls, *spectrograms, **kwargs
    ):  # pylint: disable=too-many-branches
        """Calculate a new `SpectralVariance` from a
        :class:`~gwpy.spectrogram.Spectrogram`.

        Parameters
        ----------
        spectrogram : `~gwpy.spectrogram.Spectrogram`
            input `Spectrogram` data

        bins : `~np.ndarray`, optional
            array of histogram bin edges, including the rightmost edge

        low : `float`, optional
            left edge of lowest amplitude bin, only read
            if ``bins`` is not given

        high : `float`, optional
            right edge of highest amplitude bin, only read
            if ``bins`` is not given

        nbins : `int`, optional
            number of bins to generate, only read if ``bins`` is not
            given, default: `500`

        log : `bool`, optional
            calculate amplitude bins over a logarithmic scale, only
            read if ``bins`` is not given, default: `False`

        norm : `bool`, optional
            normalise bin counts to a unit sum, default: `False`

        density : `bool`, optional
            normalise bin counts to a unit integral, default: `False`

        Returns
        -------
        specvar : `SpectralVariance`
            2D-array of spectral frequency-amplitude counts

        See also
        --------
        np.histogram
            The histogram function
        """
        # parse args and kwargs
        if not spectrograms:
            raise ValueError("Must give at least one Spectrogram")
        bins = kwargs.pop("bins", None)
        low = kwargs.pop("low", None)
        high = kwargs.pop("high", None)
        nbins = kwargs.pop("nbins", 500)
        log = kwargs.pop("log", False)
        norm = kwargs.pop("norm", False)
        density = kwargs.pop("density", False)
        if norm and density:
            raise ValueError(
                "Cannot give both norm=True and density=True, please pick one"
            )

        # get data and bins
        spectrogram = spectrograms[0]
        data = np.vstack([s.value for s in spectrograms])
        if bins is None:
            if low is None and log:
                low = np.log10(data.min() / 2)
            elif low is None:
                low = data.min() / 2
            elif log:
                low = np.log10(low)
            if high is None and log:
                high = np.log10(data.max() * 2)
            elif high is None:
                high = data.max() * 2
            elif log:
                high = np.log10(high)
            if log:
                bins = np.logspace(low, high, num=nbins + 1)
            else:
                bins = np.linspace(low, high, num=nbins + 1)
        nbins = bins.size - 1
        qbins = bins * spectrogram.unit

        # loop over frequencies
        out = np.zeros((data.shape[1], nbins))
        for i in range(data.shape[1]):
            out[i, :], bins = np.histogram(data[:, i], bins, density=density)
            if norm and out[i, :].sum():  # normalise
                out[i, :] /= out[i, :].sum()

        # return SpectralVariance
        name = f"{spectrogram.name} variance"
        try:
            new = cls(
                out,
                qbins,
                epoch=spectrogram.epoch,
                name=name,
                channel=spectrogram.channel,
                f0=spectrogram.f0,
                df=spectrogram.df,
            )
        except AttributeError:
            new = cls(
                out,
                qbins,
                epoch=spectrogram.epoch,
                name=name,
                channel=spectrogram.channel,
                f0=spectrogram.f0,
                frequencies=spectrogram.frequencies,
            )
        return new



"""
| Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""

from ._csd_daniell import (
    daniell,
    AveragingParameters,
    daniell_average,
    daniell_rearrange_fft,
)
from ._lpsd import lpsd



"""
============
Daniell averaging method for CSD
============

| Modified for Spicypy by Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Reviewed by Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>

Original description:

An upgrade to Pooya Saffarieh's, and Sam Scherf's asd2 port to Python.
This attemts to adress the following main issue, make 'csd_daniell' look like
'scipy.signal.csd' - with the goal of adding 'csd_daniell' to the
'scipy.signal' toolbox. This should offload the code maintanence to a
community with sufficient time and skill to do this.

Addresses:
    - Kwarg pass through
    - Axes
    - CSD vs ASD, CSD is ALWAYS more versatile
    - Harmonising with Scipy Library

Sources:
    - https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.csd.html
    - https://github.com/scipy/scipy/blob/v1.8.0/scipy/signal/_spectral_py.py #L454-L603
    - https://github.com/scipy/scipy/blob/v1.8.0/scipy/signal/_spectral_py.py #L1579-L1869
    - https://gitlab.nikhef.nl/6d/6dscripts/-/blob/master/SpectralAnalyzer/asd2.py
    - https://git.ligo.org/sam.scherf/ftools/-/blob/master/ftools/daniell.py
    - https://git.ligo.org/sam.scherf/ftools/-/blob/master/ftools/_utils.py

| Authors: Nathan A. Holland, Pooya Saffarieh, Abhinav Petra, Conor Mow-Lowry
| Associated Authors: Sam Scherf
| Contact: nholland@nikhef.nl
| Date: 2022-04-12
"""
from warnings import warn

import numpy as np
from gwpy.signal.spectral._utils import scale_timeseries_unit
from scipy.fft import fftshift, ifftshift  # pylint: disable=no-name-in-module
from scipy.signal import csd
from scipy.signal import get_window

from spicypy.signal.frequency_series import FrequencySeries


def daniell(*args, **kwargs):  # pylint: disable=too-many-branches
    """Calculate the CSD of a `TimeSeries` using Daniell's method.
    Frequency domain averaging of the CSD.

    Parameters
    ----------
    args : `list`
        timeseries: `TimeSeries`
            input time series (if only one specified, calculates PSD)
        other: `TimeSeries`, optional
            second input time series
        nfft: `int`
            number of samples per FFT, only value equal to full time series length is accepted
    kwargs : `dict`
        overlap : `float`, optional
            number of seconds of overlap between FFTs. User-specified value ignored, algorithm uses no overlap.
        fftlength : `float`, optional
            number of seconds in single FFT. User-specified value ignored, algorithm uses full duration.
        window : `str`, `numpy.ndarray`, optional
            Window function to apply to timeseries prior to FFT. See :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.
        number_averages: `int`, optional
            Window size for averaging (number of bins) in frequency domain. Must be an odd number. If "binning" is specified to be "logarithmic", window size will depend on bin number as `number_averages + floor(log_base ** bin_number)`
        binning: `str`, optional
            Either "constant" for constant averaging window size = `number_averages` (default) or "logarithmic" for window size depending on bin number as `number_averages + floor(log_base ** bin_number)`
        log_base: `float`, optional
            Log base for logarithmic averaging window size. Default 1.05.
        average: `str`, optional
            Averaging function to use, either "mean" (default) or "median".
        scaling: `str`, optional
            Scaling of the CSD, either "density" (default) or "spectrum".
        additional arguments, passed to scipy.signal.CSD

    Returns
    -------
    fs: FrequencySeries
        resulting CSD
    """

    try:
        timeseries, nfft = args
        other = timeseries
    except ValueError:
        timeseries, other, nfft = args

    # Convert the inputs to numpy arrays.
    x = np.asarray(timeseries.value)
    y = np.asarray(other.value)

    if len(x) != len(y):
        raise ValueError("Time series must have the same length (number of samples)!")

    # No nfft for Daniell method.
    if nfft != len(x):
        raise ValueError(
            "nfft argument is not supported by Daniell averaging method; "
            "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
        )

    # extract the rest of parameters
    sample_rate = timeseries.sample_rate.decompose().value
    ap = AveragingParameters(sample_rate, len(x), **kwargs)

    # Take the CSD.
    # Averaging isn't done by 'csd'.
    # In general the frequency vector has the ordering:
    #  0, f_positive, f_negative
    # and set noverlap to 0 (no overlap with Daniell method)
    noverlap = 0
    nperseg = len(x)  # full length of time series used for the (only) segment
    frq, CSD = csd(
        x,
        y,
        fs=sample_rate,
        window=ap.window,
        nperseg=nperseg,
        noverlap=noverlap,
        nfft=ap.nfft,
        **ap.kwargs,
    )
    # scipy returns complex values with zero imaginary part... cast it to real
    if not np.iscomplex(CSD).any():
        CSD = np.real(CSD)

    # Perform Daniell's averaging.
    frq, CSD, ap = daniell_rearrange_fft(frq, CSD, ap)
    return daniell_average(frq, CSD, timeseries, ap)


def daniell_average(frq, CSD, timeseries, ap):
    """Frequency domain averaging of the CSD.

    Parameters
    ----------
    frq : `np.array` of `float`
        frequency array
    CSD : `np.array` of `float`
        powers array
    timeseries: `TimeSeries`
        one (any of the two) of the original time series from which CSD was calculated. Needed to extract optional information such as channel name etc.
    ap: `AveragingParameters`
        parameters for averaging

    Returns
    -------
    fs: FrequencySeries
        resulting CSD
    """
    # start with CSD itself (potentially need to treat complex numbers)
    if np.iscomplexobj(CSD):
        CSD_averaged = _average(CSD.real, ap) + 1j * _average(CSD.imag, ap)
    else:  # Easy for real valued.
        CSD_averaged = _average(CSD, ap)

    # follow up with the rest: averaging frequency array and rearranging the fft order back, see `daniell_rearrange_fft`
    frq_ap = ap
    frq_ap.avg_func = np.median  # always use np.median for averaging frequency values
    frq_averaged = _average(frq, frq_ap)
    if ap.is_onesided:
        # fix DC bin - frequency equals 0 for it, not average frequency corresponding to DC bin in original CSD
        frq_averaged[0] = 0.0
    else:
        # rearrange fft back
        CSD_averaged = ifftshift(CSD_averaged)
        frq_averaged = ifftshift(frq_averaged)

    # Correct for bias.
    CSD_averaged /= ap.bias_correction(CSD)
    CSD_averaged *= ap.scale

    # generate FrequencySeries and return
    unit = scale_timeseries_unit(timeseries.unit, ap.scaling)
    return FrequencySeries(
        CSD_averaged,
        unit=unit,
        frequencies=frq_averaged,
        name=timeseries.name,
        epoch=timeseries.epoch,
        channel=timeseries.channel,
    )


def _average(CSD, ap):
    """Average CSD, helper function that is called multiple times

    Parameters
    ----------
    CSD : `np.array` of `float`
        input CSD
    ap: `AveragingParameters`
        parameters for averaging

    Returns
    -------
    CSD: `np.array` of `float`
        resulting CSD
    """
    left_edge = 0
    edges, _ = ap.window_edges(CSD)
    for right_edge in edges:
        averaged_value = ap.avg_func(CSD[left_edge:right_edge])
        if left_edge == 0:  # first bin, create CSD
            CSD_averaged = np.array(averaged_value)
        else:
            CSD_averaged = np.append(CSD_averaged, averaged_value)
        left_edge = right_edge
    return CSD_averaged


def daniell_rearrange_fft(frq, CSD, ap):
    """Rearrange FFT/CSD according to daniell method

    Parameters
    ----------
    frq : `np.array` of `float`
        frequency array
    CSD : `np.array` of `float`
        input CSD
    ap: `AveragingParameters`
        parameters for averaging

    Returns
    -------
    frq : `np.array` of `float`
        modified frequency array
    CSD : `np.array` of `float`
        modified CSD
    ap: `AveragingParameters`
        modified parameters for averaging (DC_edge, is_onesided calculated here)
    """
    ap.is_onesided = not np.any(np.less(frq, 0))
    if ap.is_onesided:
        start_drop = ap.number_averages // 2 + 1
        ap.DC_edge = start_drop
        end_drop = -1 * (frq[start_drop:].size % ap.number_averages)
        if end_drop == 0:
            end_drop = None

        frq = frq[:end_drop]
        CSD = CSD[:end_drop]
    else:
        # rearrange: put zero frequency in the middle of spectrum, between negative and positive frequencies
        frq = fftshift(frq)
        CSD = fftshift(CSD)

        # Deal with even case.
        if (frq.size % 2) == 0:
            # Cut off the lowest frequency point, which has no
            # symmetric, positive counterpart.
            frq = frq[1:]
            CSD = CSD[1:]

        total_to_drop = frq.size % ap.number_averages

        end_drop = -1 * (total_to_drop // 2)
        if end_drop == 0:
            end_drop = None

        start_drop = total_to_drop // 2
        if (total_to_drop % 2) != 0:
            # Need to drop the lowest frequency point too.
            start_drop += 1

        frq = frq[start_drop:end_drop]
        CSD = CSD[start_drop:end_drop]

    return frq, CSD, ap


class AveragingParameters:  # pylint: disable=too-many-instance-attributes
    """
    Class holding averaging parameters
    """

    def __init__(
        self, sample_rate, n_samples, **kwargs
    ):  # pylint: disable=too-many-branches
        """Constructor takes sample rate, number of samples and other optional arguments

        Parameters
        ----------
        sample_rate : int
            sampling rate of input time series
        n_samples : int
            length of time series (number of samples)
        kwargs : `dict`
            overlap : `float`, optional
                number of seconds of overlap between FFTs. User-specified value ignored, algorithm uses no overlap.
            fftlength : `float`, optional
                number of seconds in single FFT. User-specified value ignored, algorithm uses full duration.
            window : `str`, `numpy.ndarray`, optional
                Window function to apply to timeseries prior to FFT. See :func:`scipy.signal.get_window` for details on acceptable formats. Defaults to 'hann'.
            number_averages: `int`, optional
                Window size for averaging (number of bins) in frequency domain. Must be an odd number. If "binning" is specified to be "logarithmic", window size will depend on bin number as `number_averages + floor(log_base ** bin_number)`
            binning: `str`, optional
                Either "constant" for constant averaging window size = `number_averages` (default) or "logarithmic" for window size depending on bin number as `number_averages + floor(log_base ** bin_number)`
            log_base: `float`, optional
                Log base for logarithmic averaging window size. Default 1.05.
            average: `str`, optional
                Averaging function to use, either "mean" (default) or "median".
            scaling: `str`, optional
                Scaling of the CSD, either "density" (default) or "spectrum".
            additional arguments, passed to scipy.signal.CSD
        """
        # No overlap for Daniell method.
        overlap = kwargs.pop("overlap", 0)
        if overlap is not None and overlap != 0:
            raise ValueError(
                "overlap argument is not supported by Daniell averaging method; "
                "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
            )
        # noverlap is specified by GWpy by default, get rid of it
        if "noverlap" in kwargs:
            kwargs.pop("noverlap")

        # No fftlength for Daniell method.
        if kwargs.get("fftlength"):
            raise ValueError(
                "fftlength argument is not supported by Daniell averaging method; "
                "single FFT covering the full duration is calculated, with averaging performed in frequency domain."
            )
        nfft = int(n_samples + kwargs.pop("pad_length", 0))

        # get window function
        window = kwargs.pop("window_", None)
        window = "hann" if window is None else window
        # clean up default value from kwargs
        if "window" in kwargs:
            kwargs.pop("window")
        if not isinstance(window, str):
            warn(
                "Specifying window as an array for Daniell averaging method is not supported, defaulting to 'hann' window"
            )
            window = "hann"
        win = get_window(window, n_samples)

        # For now REQUIRES an odd number of averages, to handle complex
        # data types.
        number_averages = kwargs.pop("number_averages", 9)
        if (number_averages % 2) == 0:
            err = (
                f"Only an odd number of averages, not {number_averages} are "
                + "implemented - for complex symmetry reasons."
            )
            raise NotImplementedError(err)

        if int(number_averages) != number_averages or number_averages < 1:
            err = f"Number of averages must be an odd integer greater or equal 1 (currently set to {number_averages})"
            raise ValueError(err)

        binning = kwargs.pop("binning", "constant")
        if binning == "constant" and "log_base" in kwargs:
            warn(
                "`log_base` is specified but binning is set to 'constant'. "
                "Specified `log_base` value will be ignored!"
            )
        log_base = float(kwargs.pop("log_base", 1.05))

        # Check the averaging method.
        average = kwargs.pop("average", "mean")
        if average == "mean":
            avg_func = np.mean
        elif average == "median":
            avg_func = np.median
        else:
            err = (
                "Available options for <average> are 'mean' and 'median'"
                + f" not {average}."
            )
            raise ValueError(err)

        # scale as spectral density or spectrum
        scaling = kwargs.pop("scaling", "density")
        if scaling == "density":
            scale = 1.0  # scipy.signal.csd already scales for density
        elif scaling == "spectrum":
            scale = (sample_rate * (win * win).sum()) / win.sum() ** 2
        else:
            err = (
                f'Unknown scaling: {scaling}, only "density" and'
                + ' "spectrum" are supported.'
            )
            raise ValueError(err)

        # set attributes from arguments above
        self.average = average
        self.window = window
        self.win = win
        self.scale = scale
        self.scaling = scaling
        self.sample_rate = sample_rate
        self.number_averages = number_averages
        self.binning = binning
        self.log_base = log_base
        self.avg_func = avg_func
        self.nfft = nfft
        self.kwargs = kwargs

        # set defaults for other attributes
        self.DC_edge = 0
        self._bias_correction = None
        self.is_onesided = True

    def window_edges(self, CSD):
        """Calculate window edges (array indices) for each averaging bin

        Parameters
        ----------
        CSD : `np.array` of `float`
            input CSD

        Returns
        -------
        edges : `list` of `int`
            window edges (array indices) for each averaging bin
        bin_widths : `list` of `int`
            size (width) of each averaging bin
        """
        p = 0
        startpoint = 0
        endpoint = 0
        edges = []
        bin_widths = []
        if self.DC_edge != 0:  # we have a DC bin
            edges.append(self.DC_edge)
            startpoint += self.DC_edge
            bin_widths.append(self.DC_edge)

        if startpoint + self.number_averages > CSD.size:
            raise ValueError(
                "Cannot do averaging because CSD contains less elements than number_averages (+ DC bin). "
                "Try to reduce number_averages parameter to a smaller value."
            )

        while endpoint < CSD.size:
            # get right edge of current window
            step = int(self.number_averages + np.floor(self.log_base**p - 1.0))
            endpoint = startpoint + step
            edges.append(endpoint)
            bin_widths.append(endpoint - startpoint)
            startpoint += step
            if (
                self.binning == "logarithmic"
            ):  # in log case, each next window is number_averages + log_base ** p
                p = p + 1
        # fix last window edge to be as many elements as we actually have left
        if edges[-1] != CSD.size:
            edges[-1] = CSD.size
        return edges, bin_widths

    def bias_correction(self, CSD):
        """Calculate correction for median bias for each bin (depends on bin width)

        Parameters
        ----------
        CSD : `np.array` of `float`
            input CSD

        Returns
        -------
        bias_correction : `list` of `float` or `float`
            either list of float for different correction in each bin, or a single float for same correction in every bin.
        """
        if self._bias_correction is not None:
            # bias correction is already calculated, or perhaps set externally
            return self._bias_correction
        bias_correction = 1.0
        if self.average == "median":
            _, bin_widths = self.window_edges(CSD)
            bias_correction = []
            for width in bin_widths:
                if width == 1:  # no averaging in this case, hence no bias
                    correction = 1.0
                else:
                    # bias correction from _median_bias in Scipy 1.8
                    ii_2 = 2 * np.arange(1.0, (width - 1) // 2 + 1)
                    correction = 1 + np.sum(1.0 / (ii_2 + 1) - 1.0 / ii_2)
                bias_correction.append(correction)
            bias_correction = np.array(bias_correction)
        return bias_correction

    def set_bias_correction(self, correction):
        """Manually set bias correction

        Parameters
        ----------
        correction : `list` of `float` or `float`
            either list of float for different correction in each bin, or a single float for same correction in every bin.
        """
        self._bias_correction = correction



"""
A collection of useful functions for controls.

| Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Nils Leander Weickhardt <leander[dot]weickhardt[at]physikDOTuni-hamburg.de>

"""

from warnings import warn
import numpy as np
from scipy.signal import zpk2tf, ellip
from control import tf, frequency_response


def pairQ(f0, Q):
    """
    Parameters
    ----------
    f0  :   int/float
        Resonance Frequency
    Q   :   int/float
        Quality factor at f0

    Returns
    -------
    r1, r2  :   list of float
        List of two roots
    """

    # Check for correct input types
    if not isinstance(f0, (int, float)):
        raise ValueError("f0 has to be int or float")

    if not isinstance(Q, (int, float)):
        raise ValueError("Q has to be int or float")

    Q += 0j

    # resonant freq in rads/sec
    r0 = 2 * np.pi * f0

    res = np.sqrt(1 - 4 * Q**2)
    mag = r0 / (2 * Q)

    r1 = mag * (1 + res) / (2 * np.pi)
    r2 = mag * (1 - res) / (2 * np.pi)

    return [r1, r2]


def zpkgf(zeroes, poles, gain, gfreq):
    """Generate Virgo-style zpk system

    Parameters
    ----------
    zeroes  :   list of int/float/complex
        List of zeroes
    poles   :   list of int/float/complex
        List of poles
    gain    :   int/float
        System gain
    gfreq   :   int/float
        frequency at previously defined gain

    Returns
    -------
    control.TransferFunction
        Transfer function of the zpk system
    """

    # Check for correct input types
    if not isinstance(zeroes, list):
        raise ValueError("zeroes has to be a list")

    if not isinstance(poles, list):
        raise ValueError("poles has to be a list")

    if not isinstance(gain, (int, float)):
        raise ValueError("gain has to be int or float")

    if not isinstance(gfreq, (int, float)):
        raise ValueError("gfreq has to be int or float")

    # recast to numpy arrays
    zeroes = np.array(zeroes)
    poles = np.array(poles)

    pp = -2 * np.pi * poles
    zz = -2 * np.pi * zeroes

    sys = tf(*zpk2tf(zz, pp, 1))
    gain_init, _, _ = frequency_response(sys, 2 * np.pi * gfreq)
    kk = gain / np.abs(gain_init)

    return tf(*zpk2tf(zz, pp, kk))


def ellip_sus(freq, order, ripple, stopDB):
    """creates a low-pass elliptic filter with useful input parameters
    Parameters
    ----------
    freq    :   int/float
        frequency of 'knee'
    order   :   int/float
        number of poles
    ripple  :   int/float
        ripple DBs of ripple in the passband
    stopDB  :   int/float
        stopband which is stopDB DBs down

    Returns
    -------
    control.TransferFunction
        Transfer function of the elliptical filter
    """

    # Check for correct input types
    if not isinstance(freq, (int, float)):
        raise ValueError("freq has to be int or float")

    if not isinstance(order, (int, float)):
        raise ValueError("order has to be int or float")

    if not isinstance(ripple, (int, float)):
        raise ValueError("ripple has to be int or float")

    if not isinstance(stopDB, (int, float)):
        raise ValueError("stopDB has to be int or float")

    ellZ, ellP, K = ellip(order, ripple, stopDB, 1, "low", output="zpk", analog=True)
    ellZsc = 2 * np.pi * freq * ellZ
    ellPsc = 2 * np.pi * freq * ellP
    diff = len(ellP) - len(ellZ)
    Gain = (2 * np.pi * freq) ** diff

    return tf(
        *zpk2tf(ellZsc, ellPsc, K * Gain)
    )  # , ellZsc / (-2 * np.pi), ellPsc / (-2 * np.pi), K * Gain


def ellip_sus_z(freq, order, ripple, stopDB, zQ):
    """creates a derated low-pass elliptic filter with useful input parameters
    Parameters
    ----------
    freq    :   int/float
        frequency of 'knee'
    order   :   int/float
        number of poles
    ripple  :   int/float
        ripple DBs of ripple in the passband
    stopDB  :   int/float
        stopband which is stopDB DBs down
    zQ      :   int/float
        except the zeros are lower Q

    Returns
    -------
    control.TransferFunction
        Transfer function of the derated elliptical filter
    """

    # Check for correct input types
    if not isinstance(freq, (int, float)):
        raise ValueError("freq has to be int or float")

    if not isinstance(order, (int, float)):
        raise ValueError("order has to be int or float")

    if not isinstance(ripple, (int, float)):
        raise ValueError("ripple has to be int or float")

    if not isinstance(stopDB, (int, float)):
        raise ValueError("stopDB has to be int or float")

    if not isinstance(zQ, (int, float)):
        raise ValueError("zQ has to be int or float")

    if order < 2 or order > 3:
        warn("ellip_z only works for order 2 and 3")
        return ellip_sus(freq, order, ripple, stopDB)
    else:
        ellZ, ellP, K = ellip(
            order, ripple, stopDB, 1, "low", output="zpk", analog=True
        )
        Zsc = -abs(2 * np.pi * freq * ellZ[0])  # freq of the zero
        ellZsc = Zsc * np.array([1 + zQ * 1j, 1 - zQ * 1j]) / np.sqrt(1 + zQ**2)
        ellPsc = 2 * np.pi * freq * ellP
        diff = len(ellP) - len(ellZ)
        Gain = (2 * np.pi * freq) ** diff
        return tf(
            *zpk2tf(ellZsc, ellPsc, K * Gain)
        )  # , ellZsc / (-2 * np.pi), ellPsc / (-2 * np.pi), K * Gain



"""
| Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""
from .time_series import TimeSeries
from .transfer_function import TransferFunction
from .frequency_series import FrequencySeries
from .spectrogram import Spectrogram

from .filter import Filter
from .wiener_filter import WienerFilter



"""
Class wrapping :obj:`gwpy.spectrogram.spectrogram.Spectrogram` from GWpy

Authors:
    | Artem Basalaev <artem[dot]basalaev[at]pm.me>
"""
import copy
import gwpy.spectrogram
import gwpy.frequencyseries
import gwpy.timeseries
import numpy as np


class Spectrogram(gwpy.spectrogram.Spectrogram):
    """
    Class wrapping :obj:`gwpy.spectrogram.spectrogram.Spectrogram` from GWpy

    """

    @classmethod
    def from_other(cls, sgram):
        """Create Spectrogram from another Spectrogram, in particular gwpy.Spectrogram

        Parameters
        ----------
        sgram : Spectrogram, gwpy.Spectrogram
            Other Spectrogram object
        Returns
        -------
        Spectrogram
            Copy Spicypy Spectrogram object
        """
        sgram_spicypy = cls(
            sgram.value,
            unit=sgram.unit,
            xindex=sgram.xindex,
            yindex=sgram.yindex,
            name=sgram.name,
            channel=sgram.channel,
        )
        sgram_spicypy.__dict__ = copy.deepcopy(sgram_spicypy.__dict__)
        return sgram_spicypy

    @classmethod
    def from_spectra(cls, *spectra, **kwargs):
        """Build a new `Spectrogram` from a list of spectra.

        Parameters
        ----------
        *spectra
            any number of `~gwpy.frequencyseries.FrequencySeries` series
        dt : `float`, `~astropy.units.Quantity`, optional
            stride between given spectra

        Returns
        -------
        Spectrogram
            a new `Spectrogram` from a vertical stacking of the spectra
            The new object takes the metadata from the first given
            `~gwpy.frequencyseries.FrequencySeries` if not given explicitly

        Notes
        -----
        Each `~gwpy.frequencyseries.FrequencySeries` passed to this
        constructor must be the same length.
        """
        data = np.vstack([s.value for s in spectra])
        spec1 = list(spectra)[0]
        if not all(s.f0 == spec1.f0 for s in spectra):
            raise ValueError("Cannot stack spectra with different f0")
        if not all(np.array_equal(s.frequencies, spec1.frequencies) for s in spectra):
            raise ValueError("Cannot stack spectra with different df")
        kwargs.setdefault("name", spec1.name)
        kwargs.setdefault("channel", spec1.channel)
        kwargs.setdefault("epoch", spec1.epoch)
        try:
            kwargs.setdefault("df", spec1.df)
            kwargs.setdefault("f0", spec1.f0)
        except AttributeError:  #  in case of irregular frequency axis, e.g. if using LPSD
            kwargs.setdefault("frequencies", spec1.frequencies)
        kwargs.setdefault("unit", spec1.unit)
        if not ("dt" in kwargs or "times" in kwargs):
            try:
                kwargs.setdefault("dt", spectra[1].epoch.gps - spec1.epoch.gps)
            except (AttributeError, IndexError) as exc:
                raise ValueError from exc(
                    "Cannot determine dt (time-spacing) for Spectrogram from inputs"
                )
        return Spectrogram(data, **kwargs)

    def variance(  # pylint: disable=too-many-positional-arguments
        self,
        bins=None,
        low=None,
        high=None,
        nbins=500,
        log=False,
        norm=False,
        density=False,
    ):
        """Calculate the `SpectralVariance` of this `Spectrogram`.
        Overriding GWpy method to make it work for LPSD averaging method.

        Parameters
        ----------
        bins : `~numpy.ndarray`, optional, default `None`
            array of histogram bin edges, including the rightmost edge
        low : `float`, optional, default: `None`
            left edge of lowest amplitude bin, only read
            if ``bins`` is not given
        high : `float`, optional, default: `None`
            right edge of highest amplitude bin, only read
            if ``bins`` is not given
        nbins : `int`, optional, default: `500`
            number of bins to generate, only read if ``bins`` is not
            given
        log : `bool`, optional, default: `False`
            calculate amplitude bins over a logarithmic scale, only
            read if ``bins`` is not given
        norm : `bool`, optional, default: `False`
            normalise bin counts to a unit sum
        density : `bool`, optional, default: `False`
            normalise bin counts to a unit integral

        Returns
        -------
        specvar : `SpectralVariance`
            2D-array of spectral frequency-amplitude counts

        See also
        --------
        numpy.histogram
            for details on specifying bins and weights
        """
        from .frequency_series import SpectralVariance

        return SpectralVariance.from_spectrogram(
            self,
            bins=bins,
            low=low,
            high=high,
            nbins=nbins,
            log=log,
            norm=norm,
            density=density,
        )

    def _wrap_function(self, *args, **kwargs):
        """Wrap the wrap function from gwpy.Spectrogram, to return Spicypy objects"""
        out = super()._wrap_function(*args, **kwargs)
        if isinstance(out, gwpy.frequencyseries.FrequencySeries):
            # convert to Spicypy object
            from spicypy.signal import FrequencySeries

            return FrequencySeries.from_other(out)
        elif isinstance(out, gwpy.timeseries.TimeSeries):
            # convert to Spicypy object
            from spicypy.signal import TimeSeries

            return TimeSeries.from_other(out)



"""
Class extending functionality of StateSpace from python control

| Artem Basalaev <artem[dot]basalaev[at]pm.me>
| Christian Darsow-Fromm <cdarsowf[at]physnet.uni-hamburg.de>
"""
import copy

import control
import numpy as np
from spicypy.signal.time_series import TimeSeries


class System(control.LinearICSystem):
    """
    Class to model control systems and their response

    """

    def __init__(self, sys, name=None):
        """Constructor takes either control.LinearICSystem, control.InputOutputSystem, control.StateSpace, control.TransferFunction object, which is then converted to control.InterconnectedSystem, which is held by this System class.

        Parameters
        ----------
        sys : System, control.LinearICSystem, control.InputOutputSystem, control.StateSpace, control.TransferFunction
            input control system
        """
        # be careful to preserve desired name through conversions
        if name is None:
            name = sys.name

        if isinstance(sys, System):
            # already have spicypy.System, just assign it to self
            self.__dict__ = copy.deepcopy(sys.__dict__)
            self.name = name
            return

        if isinstance(sys, control.LinearICSystem):
            # have instance of parent class, assign everything to self
            self.__dict__ = copy.deepcopy(sys.__dict__)
            self.name = name
            return

        # try to convert system to LinearICSystem, in several stages
        sys_orig = (
            sys  # keep original system to show in error at the end if conversion failed
        )

        # first deal with TransferFunction, convert to StateSpace
        if isinstance(sys, control.TransferFunction):
            sys = control.tf2ss(sys)
            sys.name = name

        # then deal with StateSpace, convert to InputOutputSystem
        ss_sys = None  # save StateSpace if we got one, helps construct LinearICSystem properly
        if isinstance(sys, control.StateSpace):
            ss_sys = sys
            sys = control.ss2io(sys)
            sys.name = name

        # finally, with InputOutputSystem, convert to LinearICSystem
        if isinstance(sys, control.InputOutputSystem):
            inplist = list(sys.input_index.keys())
            for i in range(len(inplist)):
                inplist[i] = sys.name + "." + inplist[i]
            outlist = list(sys.output_index.keys())
            for i in range(len(outlist)):
                outlist[i] = sys.name + "." + outlist[i]
            ic_sys = control.InterconnectedSystem(
                [sys], inplist=inplist, outlist=outlist
            )
            super().__init__(ic_sys, ss_sys=ss_sys)
            self.name = name
            return

        # if reached here, that means all possibilities of conversion exhausted, throw an error
        raise ValueError(
            "Could not convert object", sys_orig, "into control.LinearICSystem"
        )

    def response(self, time_series, *args, **kwargs):
        """Calculate system's response to an input signal.

        Parameters
        ----------
        time_series : TimeSeries, list of TimeSeries
            input signal time series (or list for MIMO systems)

        Returns
        -------
        time_series : TimeSeries, list of TimeSeries
            output signal time series (or list for MIMO systems)
        """
        if isinstance(time_series, TimeSeries):
            t = np.array(time_series.times)
            v = np.array(time_series.value)
            resp = control.input_output_response(self, t, v, *args, **kwargs)
            return TimeSeries(resp.outputs, times=resp.time)

        if not isinstance(time_series, list):
            raise ValueError(
                "Only TimeSeries or list of TimeSeries is allowed as input signal"
            )
        if len(time_series) != self.ninputs:
            raise ValueError(
                "Length of the input list does not match number of inputs! "
                "For a MIMO system, a signal must be assigned to each input, as a list of TimeSeries. "
                "For no/empty inputs, you can specify `None` as some elements of this list."
            )

        # check that there's non-None time series on input
        ref_ts = None
        for ts in time_series:
            if ts is not None and isinstance(ts, TimeSeries):
                ref_ts = ts
        if ref_ts is None:
            raise ValueError(
                "At least one input must be TimeSeries (others can be None for no/empty input)"
            )

        t = np.array(ref_ts.times)
        v = []
        for ts in time_series:
            if ts is None:
                ts = TimeSeries(np.zeros(len(ref_ts.times)))
                ts.times = ref_ts.times
            elif not (np.array(ts.times) == t).all():
                raise ValueError("All TimeSeries must be aligned in time")
            v.append(np.array(ts.value))
        v = np.array(v)

        resp = control.input_output_response(self, t, v, *args, **kwargs)
        ts_list = []
        for output in resp.outputs:
            ts_list.append(TimeSeries(output, times=resp.time))
        return ts_list

    def add_controller(  # pylint: disable=too-many-positional-arguments
        self, control_sys, feedin, feedout, control_names=None, negative_feedback=True
    ):  # pylint: disable=too-many-branches
        """Lets you build a Matlab-like MIMO feedback system based on control.interconnect funtion.

        Parameters
        ----------
        control_sys :   list of control.TransferFunction
            List of filters in Transfer-Function form
        feedin      :   list
            List of inputs to be connected
        feedout     :   list
            List of outputs to be connected
        control_names   :   list of str, optional
            List of names for the control.TransferFunction in control_sys (default = None)
        negative_feedback   :   bool, optional
            Select if you want negative feedback or not (default = True)

        Returns
        -------
        sys_interc  :   control.LinearICSystem
            Interconnected System
        """

        # Check if control_sys is a list of TransferFunctions
        for index, filter in enumerate(control_sys):
            if isinstance(filter, control.TransferFunction) is True:
                pass
            else:
                raise ValueError(
                    "filter", index, "is not in control.TransferFunction form"
                )

        # Check if filters are proper
        for index, filter in enumerate(control_sys):
            if len(filter.num[0][0]) > len(filter.den[0][0]):
                raise ValueError(
                    "TransferFunction",
                    index,
                    "is improper: Order of denominator has to be larger or equal to order of numerator",
                )

        # Check for control_names and adapt standard naming scheme
        if control_names is None:
            control_names = [filter.name for filter in control_sys]

        if len(control_names) != len(control_sys):
            raise ValueError(
                "control_names has to be either None or a list of the same length as control_sys"
            )

        # Check if feedin and feedout have the same dimensions as control_sys
        if len(feedin) != len(feedout) != len(control_sys):
            raise ValueError(
                "feedin, feedout and control_sys need to be the same length"
            )

        # Check sign of feedback
        if isinstance(negative_feedback, bool) is False:
            raise ValueError("negative_feedback has to be either True or False")
        feedback = ""
        if negative_feedback is True:
            feedback = "-"

        # Convert filters to LinearIOSystem and give name
        control_sys = self._controller_tf_to_io(control_sys, control_names)

        # Prepare connection matrix
        connections = []
        # Create connections
        for i, index_in, index_out in zip(
            np.arange(0, len(feedin) + 1), feedin, feedout
        ):
            connections.append(
                [
                    control_sys[i].name + ".u[0]",
                    self.name + ".y[" + str(index_in) + "]",
                ]
            )
            connections.append(
                [
                    self.name + ".u[" + str(index_out) + "]",
                    feedback + control_sys[i].name + ".y[0]",
                ]
            )

        # create sadly necessary helper bs (until control version 0.10)
        inp = []
        outp = []

        ninputs = self.ninputs
        noutputs = self.noutputs

        for i in range(ninputs):
            inp.append(self.name + ".u[" + str(i) + "]")
        for i in range(noutputs):
            outp.append(self.name + ".y[" + str(i) + "]")

        # compute interconnected system
        sys_interc = control.interconnect(
            [self, *control_sys], connections=connections, inplist=inp, outlist=outp
        )
        return System(sys_interc)

    def _controller_tf_to_io(self, control_sys, control_names):
        """Helper method to convert control.TransferFunction object into control.LinearIOSystem

        Parameters
        ----------
        control_sys :   list of control.TransferFunction
            list of control.Transferfunction to be converted into control.LinearIOSystem
        control_names   :   list of str
            list of names to be given to the control filters

        Returns
        -------
        control_sys :  list of control.LinearIOSystem
            Same as input parameter but now as list of control.LinearIOSystem
        """

        for index, filter in enumerate(control_sys):
            control_sys[index] = control.tf2io(filter, name=control_names[index])
        return control_sys

    def __add__(self, *args, **kwargs):
        """Add two LTI systems (parallel connection). See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__add__(*args, **kwargs))

    def __getitem__(self, *args, **kwargs):
        """Array style access. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__getitem__(*args, **kwargs))

    def __mul__(self, *args, **kwargs):
        """Multiply two LTI objects (serial connection). See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__mul__(*args, **kwargs))

    def __neg__(self):
        """Negate a state space system. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__neg__())

    def __radd__(self, *args, **kwargs):
        """Right add two LTI systems (parallel connection). See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__radd__(*args, **kwargs))

    def __rmul__(self, *args, **kwargs):
        """Right multiply two LTI objects (serial connection). See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__rmul__(*args, **kwargs))

    def __rsub__(self, other):
        """Right subtract two LTI systems. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__rsub__(other))

    def __sub__(self, *args, **kwargs):
        """Subtract two LTI systems. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().__sub__(*args, **kwargs))

    def __truediv__(self, *args, **kwargs):
        """Division of state space systems by TFs, FRDs, scalars, and arrays.
        See `LinearICSystem` in python-control docs. Here we only wrap it in `System` class.
        """
        return System(super().__truediv__(*args, **kwargs))

    def append(self, *args, **kwargs):
        """Append a second model to the present model. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().append(*args, **kwargs))

    def copy(self, *args, **kwargs):
        """Make a copy of an input/output system. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().copy(*args, **kwargs))

    def feedback(self, *args, **kwargs):
        """Add a feedback connection. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().feedback(*args, **kwargs))

    def linearize(self, *args, **kwargs):
        """Linearize an input/output system at a given state and input. See `LinearICSystem` in python-control docs.
        Here we only wrap it in `System` class.
        """
        return System(super().linearize(*args, **kwargs))

    def minreal(self, *args, **kwargs):
        """Calculate a minimal realization, removes unobservable and uncontrollable states.
        See `LinearICSystem` in python-control docs. Here we only wrap it in `System` class.
        """
        return System(super().minreal(*args, **kwargs))
